/************************************************************************/
/*  File:       ashtech_bin.h                                           */
/*                                                                      */
/*  Description:                                                        */
/*      definitions and structures for binary Ashtech XII data.         */
/*                                                                      */
/*      Douglas Neuhauser                                               */
/*      Seismographic Station                                           */
/*      University of California, Berkeley                              */
/*      doug@seismo.berkeley.edu                                        */
/*                                                                      */
/************************************************************************/

/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
Modification History:
------------------------------------------------------------------------
95/11/13 DSN    1.0     Initial coding.
96/06/16 DSN    1.3     Add RTR error record.
96/10/01 DSN    1.4     Added copyright msg.
*/

/************************************************************************/

/*      @(#)ashtech_bin.h       1.1 09/09/97 19:04:10   */

#define UNKNOWN_REC     -1              /* Unknown record id.           */
#define MBEN_CA_REC     0               /* MBEN CA record id.           */
#define MBEN_CL_REC     1               /* MBEN L2 codeless record id.  */
#define MBEN_P1_REC     2               /* MBEN P-code L1 record id.    */
#define MBEN_P2_REC     3               /* MBEN P-code L2 record id.    */
#define MBEN_PC_REC     4               /* MBEN P-code L1&L2 record id. */
#define PBEN_REC        5               /* PBEN record id.              */
#define SNAV_REC        6               /* SNAV record id.              */
#define SALM_REC        7               /* SALM record id.              */
#define RTR_REC         8               /* RTR record id.               */

#define MBEN_CA_LEN     50              /* MBEN CA record len.          */
#define MBEN_CL_LEN     62              /* MBEN L2 codeless record len. */
#define MBEN_P1_LEN     79              /* MBEN P-code L1 record len.   */
#define MBEN_P2_LEN     79              /* MBEN P-code L2 record len.   */
#define MBEN_PC_LEN     108             /* MBEN P-code L1&l2 record len.*/
#define PBEN_LEN        69              /* PBEN record len.             */
#define SNAV_LEN        145             /* SNAV record len.             */
#define SALM_LEN        83              /* SALM record len.             */
#define SALM_DATA_LEN   69              /* SALM data len.               */
#define RTR_LEN         21              /* RTR error record len.        */

#define L1_PCODE_BLOCK  1               /* Code for L1 P-code block.    */
#define L2_PCODE_BLOCK  2               /* Code for L2 P-code block.    */
#define CA_CODE_BLOCK   3               /* Code for CA code block.      */
#define L2_CODELESS     4               /* Code for L2 codeless block.  */

#define MAXRLEN         256             /* Max record len in chars.     */
#define MAXSLEN         128             /* Max record len in short ints.*/

#define MBEN_ID_PER_SEC 20              /* 1/seq_id units (secs).       */
#define PBEN_RES_PER_SEC 1000           /* 1/pbentime units (secs).     */

typedef struct _code_block {    /* Data from code block of MBEN record. */
    unsigned char type;         /* type of code block (C/A, P-L1, p_L2).*/
    unsigned char warning;      /* warning flag (0 == OK).              */
    unsigned char goodbad;      /* 24 == OK. < 24 -> not used.          */
    unsigned char polarity_know;/* unused in Z12 receiver.              */
    unsigned char ireg;         /* satellite snr: 0-5, 95-100 are OK.   */
    unsigned char qa_phase;     /* phase quality indicator.             */
    double full_phase;          /* full carrier phase in cycles.        */
    double raw_range;           /* raw range to sv (sec).               */
    int doppler;                /* doppler (10^-4 Hz).                  */
    int smoothing;              /* smoothing.                           */
} CODE_BLOCK;

typedef struct _codeless_block {/* Data from L2 codelss block of MBEN.  */
    unsigned char type;         /* type of codeless data (L2 codeless). */
    double carphase2;           /* full carrier phase L2 (cycles).      */
    unsigned char ireg2;        /* snr for L2.                          */
    char qa_phase2;             /* satellite snr: 0-5, 95-100 are OK.   */
} CODELESS_BLOCK;

typedef struct _mben {          /* Data from MBEN record.               */
    short type;                 /* record type.                         */
    /* Block id                                                         */
    unsigned short seq_tag;     /* seq id in msec (mod 5 min).          */
    unsigned char left;         /* # remaining mben in epoch.           */
    /* Satellite data                                                   */
    unsigned char svprn;        /* satellite PRN number.                */
    unsigned char elevation;    /* 1 degree per unit.                   */
    unsigned char azimuth;      /* 2 degrees per unit.                  */
    unsigned char chnind;       /* channel ID (1 - 12).                 */
    /* Code blocks for C/A, P code on L1, P code on L2.                 */
    CODE_BLOCK code[3];         /* code blocks.                         */
    /* L2 codeless block.                                               */
    CODELESS_BLOCK codeless;    /* codeless L2 data.                    */
} MBEN;

typedef struct _pben {          /* Data from PBEN record.               */
    short type;                 /* record type.                         */
    int pbentime;               /* receive time of data (10^-3 sec).    */
    char sitename[5];           /* 4-character sitename.                */
    double navx;                /* station position: ECEF-X.            */
    double navy;                /* station position: ECEF-Y.            */
    double navz;                /* station position: ECEF-Z.            */
    float navt;                 /* clock offset (meters).               */
    float navxdot;              /* velocity in ECEF-X (m/sec).          */
    float navydot;              /* velocity in ECEF-Y (m/sec).          */
    float navzdot;              /* velocity in ECEF-Z (m/sec).          */
    float navtdot;              /* clock drift.                         */
    unsigned short pdop;        /* PDOP.                                */
} PBEN;

typedef struct _snav {          /* Data from SNAV record.               */
    short type;                 /* record type (SNAV)                   */
    short wn;                   /* GPS week number.                     */
    int tow;                    /* seconds of GPS week.                 */
    float tgd;                  /* group delay (sec).                   */
    int aodc;                   /* clock data issue.                    */
    int toc;                    /* (sec).                               */
    float af2;                  /* clock: (sec/sec^2).                  */
    float af1;                  /* clock: (sec/sec).                    */
    float af0;                  /* clock: (sec).                        */
    int aode;                   /* orbit data issue.                    */
    float deltan;               /* mean anomaly corr. (semi-circle/sec).*/
    double m0;                  /* mean anomaly @ reftime (semi-circle).*/
    double e;                   /* eccentricity.                        */
    double roota;               /* sqrt of semi-major axis (meters 1/2).*/
    int toe;                    /* ref time for orbit (sec).            */
    float cic;                  /* harmonic correction term (radians).  */
    float crc;                  /* harmonic correction term (meters).   */
    float cis;                  /* harmonic correction term (radians).  */
    float crs;                  /* harmonic correction term (meters).   */
    float cuc;                  /* harmonic correction term (radians).  */
    float cus;                  /* harmonic correction term (meters).   */
    double omega0;              /* Lon of Asc. node (semi-circles).     */
    double omega;               /* Arg of Perigee (semi-circles).       */
    double i0;                  /* incl. angle @ reftime (semi-circles).*/
    float omegadot;             /* rate of right asc. (semi-circles/sec)*/
    float idot;                 /* rate of incl (semi-circles/sec).     */
    short int accuracy;         /* (coded).                             */
    short int health;           /* (coded).                             */
    short int fit;              /* curt fit interval (coded).           */
    char prnnum;                /* SV PRN number - 1 (Z12 receiver).    */
} SNAV;

typedef struct _salm {          /* Data from SALM structure - unknown.  */
    char private[SALM_DATA_LEN];/* private data.                        */
} SALM;

typedef struct _rtr {           /* Data from RTR error record.          */
    char hex[8];                /* ASCII hex representation of 32-bits. */
} RTR;

union u_bin {                   /* Union for all record types.          */
    MBEN mben;                  /* MBEN record type.                    */
    PBEN pben;                  /* PBEN record type.                    */
    SNAV snav;                  /* SNAV record type.                    */
    SALM salm;                  /* SALM record type.                    */
    RTR rtr;                    /* RTR error record type.               */
} u;
