/************************************************************************/
/*  Interface routines for Ashtech XII GPS receiver.                    */
/*                                                                      */
/*      Douglas Neuhauser                                               */
/*      Seismographic Station                                           */
/*      University of California, Berkeley                              */
/*      doug@seismo.berkeley.edu                                        */
/*                                                                      */
/************************************************************************/

/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/************************************************************************/

#ifndef lint
static char sccsid[] = "@(#)ashtech_subs.c      1.1 09/09/97 19:03:42";
#endif	/* lint */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

#ifndef _OSK
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <qlib2.h>
#else	/* _OSK */
#include <types.h>
#include <sysio.h>
#define strrchr rindex
/* time.h included by stdlib.h */
#endif	/* _OSK */

#include "gps_generic.h"
#include "receiver_subs.h"
#include "serial_io.h"
#include "ashtech_bin.h"

/************************************************************************/
/*  Vendor-specific definitions.                                        */
/************************************************************************/
#define DEFAULT_PATTERN "%s%j0.r%H"
char gps_vendor[120]   = "ASHTECH";
char gps_model[120]    = "Z-12";
char gps_software[120] = "";
char gps_firmware[120] = "";

/************************************************************************/
/*  Externals                                                           */
/************************************************************************/
/*                                                                      */
/*  Allocated buffer for use in main procecure, since the size of the   */
/*  records is receiver dependent.                                      */
/************************************************************************/
char input_rec[MAXRLEN];        /* buffer to hold input binary records. */
int maxrlen = MAXRLEN;          /* max input record length.             */
char out_pattern[] = DEFAULT_PATTERN;   /* outfile pattern template.    */
extern FILE *info;              /* output file ptr for information.     */
extern int debug_option;        /* default debug option.                */
extern int verbosity;           /* verbosity option.                    */
#ifndef _OSK
extern FILE *out;               /* output file ptr.                     */
#else	/* _OSK */
extern int out;                 /* output file descriptor.              */
#endif	/* _OSK */

/************************************************************************/
/*  Static variables.                                                   */
/************************************************************************/
static int decimate;
static int sample_interval;

static int mben_interval;
static int pben_interval;

static char CR = 0x0d;
static char NL = 0x0a;

/************************************************************************/
/*  Internal function declarations.                                     */
/************************************************************************/
int parse_codeless_data (int type, unsigned char *line, CODELESS_BLOCK *cb);
int parse_code_data (int type, unsigned char *line, CODE_BLOCK *cb);
int parse_mben (int type, unsigned char *line, int len, MBEN *mben);
int parse_pben (int type, unsigned char *line, int len, PBEN *pben);
int parse_snav (int type, unsigned char *line, int len, SNAV *snav);
int parse_salm (int type, unsigned char *line, int len, SALM *salm);

int dump_codeless_data (int type, char *prefix, CODELESS_BLOCK *cb, FILE *fp);
int dump_code_data (int type, char *prefix, CODE_BLOCK *cb, FILE *fp);
int dump_mben (int type, char *prefix, MBEN *mben, FILE *fp);
int dump_pben (int type, char *prefix, PBEN *pben, FILE *fp);
int dump_snav (int type, char *prefix, SNAV *snav, FILE *fp);
int dump_salm (int type, char *prefix, SALM *salm, FILE *fp);

/************************************************************************/
/*  setup_decimation:                                                   */
/*      Setup for decimation.                                           */
/*  Return:                                                             */
/*      0 on OK, -1 on error.                                           */
/************************************************************************/
int setup_decimation (int decimate_flag, int sample_interval_val)
{
    decimate = decimate_flag;
    sample_interval = sample_interval_val;
    if (decimate && sample_interval <= 0.0) {
        fprintf (stderr, "Error - sample interval must be > 0\n");
        return (-1);
    }
    if (decimate) {
        mben_interval = MBEN_ID_PER_SEC * sample_interval;
        pben_interval = PBEN_RES_PER_SEC * sample_interval;
    }
    return (0);
}
    
/************************************************************************/
/*  get_bin_record -                                                    */
/*      read, verify, and determine type of binary record.              */
/************************************************************************/
int get_bin_record (char *line, int maxlen, int *rec_type, int *rec_len, int *nskip, int fd)
{
    int c;
    int i;
    static unsigned char fixed_hdr[] = "$PASHR,";
    int fixed_hdr_len = strlen((void *)fixed_hdr);
    int reclen;
    int cksum;
    int nc, ns;
    unsigned short tmp[MAXSLEN];
    unsigned short ckshort;

    /* Frame on header  */
    if (verbosity >= 1) fprintf (info, "Enter get_bin_record\n");
    *nskip = 0;
    while (1) {
        *rec_len = 0;
        *rec_type = UNKNOWN_REC;
        while ((c=nextc(fd)) != '$') {
            if (c==EOF) {
                if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n",
                                             EOF);
                return(EOF);
            }
            ++*nskip;
        }
        line[0] = c;
        ++*rec_len;
        for (i=1; i<fixed_hdr_len; i++) {
            c = nextc(fd);
            if (c==EOF) {
                if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n",
                                             EOF);
                return(EOF);
            }
            if (c != fixed_hdr[i]) continue;
            line[i] = c;
            ++*rec_len;
        }
        /* Scan for fixed_hdr again if we have not seen all of it.  */
        if (i != fixed_hdr_len) {
            *nskip += i+1;
            continue;
        }
        /* Read the record_type portion of the header.              */
        for (i=fixed_hdr_len; i<fixed_hdr_len+4; i++) {
            c = nextc(fd);
            if (c==EOF) {
                if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n",
                                             EOF);
                return(EOF);
            }
            line[i] = c;
            ++*rec_len;
        }
        break;
    }
    if (DEBUG(DEBUG_INPUT)) {
        fprintf (info, "Header found\n");
    }
    /* Entire header has been read.  Determine type of hdr,         */
    /* and therefore the record length.                             */
    if (strncmp((char *)(line+fixed_hdr_len),"MCA",3)==0) {
        *rec_type = MBEN_CA_REC;
        reclen = MBEN_CA_LEN;
    }
    else if (strncmp((char *)(line+fixed_hdr_len),"MCL",3)==0) {
        *rec_type = MBEN_CL_REC;
        reclen = MBEN_CL_LEN;
    }
    else if (strncmp((char *)(line+fixed_hdr_len),"MP1",3)==0) {
        *rec_type = MBEN_P1_REC;
        reclen = MBEN_P1_LEN;
    }
    else if (strncmp((char *)(line+fixed_hdr_len),"MP2",3)==0) {
        *rec_type = MBEN_P2_REC;
        reclen = MBEN_P2_LEN;
    }
    else if (strncmp((char *)(line+fixed_hdr_len),"MPC",3)==0) {
        *rec_type = MBEN_PC_REC;
        reclen = MBEN_PC_LEN;
    }
    else if (strncmp((char *)(line+fixed_hdr_len),"PBN",3)==0) {
        *rec_type = PBEN_REC;
        reclen = PBEN_LEN;
    }
    else if (strncmp((char *)(line+fixed_hdr_len),"SNV",3)==0) {
        *rec_type = SNAV_REC;
        reclen = SNAV_LEN;
    }
    else if (strncmp((char *)(line+fixed_hdr_len),"ALM",3)==0) {
        *rec_type = SALM_REC;
        reclen = 83;
    }
    else {
        *rec_type = UNKNOWN_REC;
        *rec_len = fixed_hdr_len+4;
        if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n",
                                     RECORD_ERROR);
        return (RECORD_ERROR);
    }

    /* Read the rest of the record.                             */
    cksum = 0;
    for (i=fixed_hdr_len+4; i < reclen; i++) {
        c = nextc(fd);
        if (c==EOF) {
            *rec_len = i;
            if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n",
                                         EOF);
            return(EOF);
        }
        line[i] = c;
        ++*rec_len;
    }
    switch (*rec_type) {
    case MBEN_CA_REC:
    case MBEN_CL_REC:
    case MBEN_P1_REC:
    case MBEN_P2_REC:
    case MBEN_PC_REC:
        /* Compute unsigned char checksum (XOR of bytes).       */
        cksum = 0;
        for (i=fixed_hdr_len+4; i<reclen-3; i++) {
            cksum ^= line[i];
        }
        if (cksum != line[reclen-3]) {
            if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n",
                                         CKSUM_ERROR);
            return (CKSUM_ERROR);
        }
        break;
    case PBEN_REC:
    case SNAV_REC:
        /* Compute short work checksum (sum or short words).    */
        /* Copy to temp buffer to ensure short alignment.       */
        nc = reclen - fixed_hdr_len - 4 - 4;
        ns = nc >> 1;
        ckshort = 0;
        memcpy ((void *)tmp, (void *)(line+fixed_hdr_len+4), nc + 2);
        for (i=0; i<ns; i++) {
            ckshort += tmp[i];
        }
        if (ckshort != tmp[ns]) {
            if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n",
                                         CKSUM_ERROR);
            return (CKSUM_ERROR);
        }
        break;
    case SALM_REC:
        /* Do not know the format of this record.               */
        break;
    }

    /* Perform integrity check on the end of record.            */
    if (! ((char)line[reclen-2] == CR && (char)line[reclen-1] == NL)) {
        if (DEBUG(DEBUG_INPUT)) {
            fprintf (info, "expected 0x%x 0x%x, got: 0x%x 0x%x\n",
                     CR, NL, line[reclen-2], line[reclen-1]);
        }
        if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n",
                                     TERM_ERROR);
        return (TERM_ERROR);
    }

    /* Return record type;                                      */
    *rec_len = reclen;
    if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=0\n");
    return (0);
}

/************************************************************************/
/*  parse_codeless_data -                                               */
/*      Parse L2 codeless data.                                         */
/************************************************************************/
int parse_codeless_data (int type, unsigned char *line, CODELESS_BLOCK *cb)
{
    cb->type = type;
    memcpy ((void *)&cb->carphase2, (void *)&line[0], 8);
    cb->ireg2 = line[8];
    cb->qa_phase2 = line[9];
    return (0);
}

/************************************************************************/
/*  dump_codeless_data -                                                */
/*      Dump L2 codeless data.                                          */
/************************************************************************/
int dump_codeless_data (int type, char *prefix, CODELESS_BLOCK *cb, FILE *fp)
{
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "type", cb->type);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "carphase2", cb->carphase2);
    fprintf (fp, "%s %-15.15s = %c\n", prefix, "ireg2", cb->ireg2);
    fprintf (fp, "%s %-15.15s = %c\n", prefix, "qa_phase2", cb->qa_phase2);
    return (0);
}

/************************************************************************/
/*  parse_code_data -                                                   */
/*      Parse C/A, P-L1 or P_L2 code data.                              */
/************************************************************************/
int parse_code_data (int type, unsigned char *line, CODE_BLOCK *cb)
{
    cb->type = type;
    cb->warning = line[0];
    cb->goodbad = line[1];
    cb->polarity_know = line[2];
    cb->ireg = line[3];
    cb->qa_phase = line[4];
    memcpy ((void *)&cb->full_phase, (void *)&line[5], 8);
    memcpy ((void *)&cb->raw_range, (void *)&line[13], 8);
    memcpy ((void *)&cb->doppler, (void *)&line[21], 4);
    memcpy ((void *)&cb->smoothing, (void *)&line[25], 4);
    return (0);
}

/************************************************************************/
/*  dump_code_data -                                                    */
/*      Dump C/A, P-L1 or P_L2 code data.                               */
/************************************************************************/
int dump_code_data (int type, char *prefix, CODE_BLOCK *cb, FILE *fp)
{
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "type", cb->type);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "warning", cb->warning);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "goodbad", cb->goodbad);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "polarity_know", cb->polarity_know);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "ireg", cb->ireg);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "qa_phase", cb->qa_phase);
/*::These fields have been removed from the current Ashtech document.
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "full_phase", cb->full_phase);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "raw_range", cb->raw_range);
  ::*/
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "doppler", cb->doppler);
    fprintf (fp, "%s %-15.15s = 0x%-8.8x (%d %d)\n", prefix, "smoothing", 
             cb->smoothing,
             (cb->smoothing>>24) & 0xFF,
             ((((cb->smoothing>>23) & 0x1) ? -1 : 1) *
              (cb->smoothing & 0x7FFFFF)));
    return (0);
}

/************************************************************************/
/*  parse_mben -                                                        */
/*      Parse MBEN record.                                              */
/************************************************************************/
int parse_mben (int type, unsigned char *line, int len, MBEN *mben)
{
    mben->type = type;
    memcpy ((void *)(&mben->seq_tag), (void *)(line+11), 2);
    mben->left = line[13];
    mben->svprn = line[14];
    mben->elevation = line[15];
    mben->azimuth = line[16];
    mben->chnind = line[17];
    switch (type) {
    case MBEN_CA_REC:
        parse_code_data (CA_CODE_BLOCK, line+18, &mben->code[0]);
        break;
    case MBEN_CL_REC:
        parse_code_data (CA_CODE_BLOCK, line+18, &mben->code[0]);
        parse_codeless_data (L2_CODELESS, line+18+29, &mben->codeless);
        break;
    case MBEN_P1_REC:
        parse_code_data (CA_CODE_BLOCK, line+18, &mben->code[0]);
        parse_code_data (L1_PCODE_BLOCK, line+29, &mben->code[1]);
        break;
    case MBEN_P2_REC:
        parse_code_data (CA_CODE_BLOCK, line+18, &mben->code[0]);
        parse_code_data (L2_PCODE_BLOCK, line+29, &mben->code[2]);
        break;
    case MBEN_PC_REC:
        parse_code_data (CA_CODE_BLOCK, line+18, &mben->code[0]);
        parse_code_data (L1_PCODE_BLOCK, line+18+29, &mben->code[1]);
        parse_code_data (L2_PCODE_BLOCK, line+18+29+29, &mben->code[2]);
        break;
    default:
#ifndef _OSK
        fprintf (stderr, "Unknown MBEN type %d\n", type);
#else	/* _OSK */
        fprintf (stderr, "Unknown MBEN type %d\n", type);
#endif	/* _OSK */
        return (FORMAT_ERROR);
    }
    return (0);
}

/************************************************************************/
/*  dump_mben -                                                         */
/*      Dump MBEN record.                                               */
/************************************************************************/
int dump_mben (int type, char *prefix, MBEN *mben, FILE *fp)
{
    char ca_prefix[80], l2_codeless_prefix[80];
    char l1_pcode_prefix[80], l2_pcode_prefix[80];
    sprintf (ca_prefix, "%s ca_code", prefix);
    sprintf (l1_pcode_prefix, "%s l1_pcode", prefix);
    sprintf (l2_pcode_prefix, "%s l2_pcode", prefix);
    sprintf (l2_codeless_prefix, "%s l2_codeless", prefix);

    fprintf (fp, "%s %-15.15s = %d\n", prefix, "type", mben->type);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "seq_tag", mben->seq_tag);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "left", mben->left);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "svprn", mben->svprn);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "elevation", mben->elevation);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "azimuth", mben->azimuth);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "chnind", mben->chnind);
    switch (type) {
    case MBEN_CA_REC:
        dump_code_data (CA_CODE_BLOCK, ca_prefix, &mben->code[0], fp);
        break;
    case MBEN_CL_REC:
        dump_code_data (CA_CODE_BLOCK, ca_prefix, &mben->code[0], fp);
        dump_codeless_data (L2_CODELESS, l2_codeless_prefix, &mben->codeless, fp);
        break;
    case MBEN_P1_REC:
        dump_code_data (CA_CODE_BLOCK, ca_prefix, &mben->code[0], fp);
        dump_code_data (L1_PCODE_BLOCK, l1_pcode_prefix, &mben->code[1], fp);
        break;
    case MBEN_P2_REC:
        dump_code_data (CA_CODE_BLOCK, ca_prefix, &mben->code[0], fp);
        dump_code_data (L2_PCODE_BLOCK, l2_pcode_prefix, &mben->code[2], fp);
        break;
    case MBEN_PC_REC:
        dump_code_data (CA_CODE_BLOCK, ca_prefix, &mben->code[0], fp);
        dump_code_data (L1_PCODE_BLOCK, l1_pcode_prefix, &mben->code[1], fp);
        dump_code_data (L2_PCODE_BLOCK, l2_pcode_prefix, &mben->code[2], fp);
        break;
    default:
#ifndef _OSK
        fprintf (fp, "Unknown MBEN type %d\n", type);
#else	/* _OSK */
        fprintf (fp, "Unknown MBEN type %d\n", type);
#endif	/* _OSK */
        return (FORMAT_ERROR);
    }
    fprintf (fp, "\n");
    return (0);
}

/************************************************************************/
/*  parse_pben -                                                        */
/*      Parse PBEN record.                                              */
/************************************************************************/
int parse_pben (int type, unsigned char *line, int len, PBEN *pben)
{
    pben->type = type;
    memcpy ((void *)&pben->pbentime, (void *)&line[11], 4);
    memcpy ((void *)&pben->sitename, (void *)&line[15], 4);
    memcpy ((void *)&pben->navx, (void *)&line[19], 8);
    memcpy ((void *)&pben->navy, (void *)&line[27], 8);
    memcpy ((void *)&pben->navz, (void *)&line[35], 8);
    memcpy ((void *)&pben->navt, (void *)&line[43], 4);
    memcpy ((void *)&pben->navxdot, (void *)&line[47], 4);
    memcpy ((void *)&pben->navydot, (void *)&line[51], 4);
    memcpy ((void *)&pben->navzdot, (void *)&line[55], 4);
    memcpy ((void *)&pben->navtdot, (void *)&line[59], 4);
    memcpy ((void *)&pben->pdop, (void *)&line[63], 2);
    return (0);
}

/************************************************************************/
/*  dump_pben -                                                         */
/*      Dump PBEN record.                                               */
/************************************************************************/
int dump_pben (int type, char *prefix, PBEN *pben, FILE *fp)
{
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "type", pben->type);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "pbentime", pben->pbentime);
    fprintf (fp, "%s %-15.15s = %s\n", prefix, "sitename", pben->sitename);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "navx", pben->navx);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "navy", pben->navy);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "navz", pben->navz);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "naft", pben->navt);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "navxdot", pben->navxdot);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "navydot", pben->navydot);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "navzdot", pben->navzdot);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "navtdot", pben->navtdot);
    fprintf (fp, "%s %-15.15s = %hu\n", prefix, "pdop", pben->pdop);
    fprintf (fp, "\n");
    return (0);
}

/************************************************************************/
/*  parse_snav -                                                        */
/*      Parse SNAV record.                                              */
/************************************************************************/
int parse_snav (int type, unsigned char *line, int len, SNAV *snav)
{
    memcpy ((void *)&snav->wn, (void *)&line[11], 2);
    memcpy ((void *)&snav->tow, (void *)&line[13], 4);
    memcpy ((void *)&snav->tgd, (void *)&line[17], 4);
    memcpy ((void *)&snav->aodc, (void *)&line[21], 4);
    memcpy ((void *)&snav->toc, (void *)&line[25], 4);
    memcpy ((void *)&snav->af2, (void *)&line[29], 4);
    memcpy ((void *)&snav->af1, (void *)&line[33], 4);
    memcpy ((void *)&snav->af0, (void *)&line[37], 4);
    memcpy ((void *)&snav->aode, (void *)&line[41], 4);
    memcpy ((void *)&snav->deltan, (void *)&line[45], 4);
    memcpy ((void *)&snav->m0, (void *)&line[49], 8);
    memcpy ((void *)&snav->e, (void *)&line[57], 8);
    memcpy ((void *)&snav->roota, (void *)&line[65], 8);
    memcpy ((void *)&snav->toe, (void *)&line[73], 4);
    memcpy ((void *)&snav->cic, (void *)&line[77], 4);
    memcpy ((void *)&snav->crc, (void *)&line[81], 4);
    memcpy ((void *)&snav->cis, (void *)&line[85], 4);
    memcpy ((void *)&snav->crs, (void *)&line[89], 4);
    memcpy ((void *)&snav->cuc, (void *)&line[93], 4);
    memcpy ((void *)&snav->cus, (void *)&line[97], 4);
    memcpy ((void *)&snav->omega0, (void *)&line[101], 8);
    memcpy ((void *)&snav->omega, (void *)&line[109], 8);
    memcpy ((void *)&snav->i0, (void *)&line[117], 8);
    memcpy ((void *)&snav->omegadot, (void *)&line[125], 4);
    memcpy ((void *)&snav->idot, (void *)&line[129], 4);
    memcpy ((void *)&snav->accuracy, (void *)&line[133], 2);
    memcpy ((void *)&snav->health, (void *)&line[135], 2);
    memcpy ((void *)&snav->fit, (void *)&line[137], 2);
    snav->prnnum = line[139];
    return (0);
}

/************************************************************************/
/*  dump_snav -                                                         */
/*      Dump SNAV record.                                               */
/************************************************************************/
int dump_snav (int type, char *prefix, SNAV *snav, FILE *fp)
{
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "type", snav->type);
    fprintf (fp, "%s %-15.15s = %hd\n", prefix, "wn", snav->wn);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "tow", snav->tow);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "tgd", snav->tgd);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "aodc", snav->aodc);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "toc", snav->toc);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "af2", snav->af2);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "af1", snav->af1);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "af0", snav->af0);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "aode", snav->aode);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "deltan", snav->deltan);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "m0", snav->m0);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "e", snav->e);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "roota", snav->roota);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "toe", snav->toe);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "cic", snav->cic);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "crc", snav->crc);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "cis", snav->cis);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "crs", snav->crs);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "cuc", snav->cuc);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "cus", snav->cus);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "omega0", snav->omega0);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "omega", snav->omega);
    fprintf (fp, "%s %-15.15s = %21.14le\n", prefix, "i0", snav->i0);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "omegadot", snav->omegadot);
    fprintf (fp, "%s %-15.15s = %f\n", prefix, "idot", snav->idot);
    fprintf (fp, "%s %-15.15s = %hd\n", prefix, "accuracy", snav->accuracy);
    fprintf (fp, "%s %-15.15s = %hd\n", prefix, "health", snav->health);
    fprintf (fp, "%s %-15.15s = %hd\n", prefix, "fit", snav->fit);
    fprintf (fp, "%s %-15.15s = %d\n", prefix, "prnnum", snav->prnnum);
    fprintf (fp, "\n");
    return (0);
}

/************************************************************************/
/*  parse_salm -                                                        */
/*      Parse SALM record.                                              */
/************************************************************************/
int parse_salm (int type, unsigned char *line, int len, SALM *salm)
{
    int maxlen = (len <= SALM_DATA_LEN) ? len : SALM_DATA_LEN;
    memcpy ((void *)salm->private, (void *)&line[11], maxlen);
    return (0);
}

/************************************************************************/
/*  dump_salm -                                                         */
/*      Dump SALM record.                                               */
/************************************************************************/
int dump_salm (int type, char *prefix, SALM *salm, FILE *fp)
{
    fprintf (fp, "%s %-15.15s = <private>\n", prefix, "contents");
    fprintf (fp, "\n");
    return (0);
}

/************************************************************************/
/*  process_record:                                                     */
/*      Process input record.                                           */
/*  Return:                                                             */
/*      Negative value on error.                                        */
/*      Non-negative value on record OK.                                */
/*      0:  output record.                                              */
/*      1:  skip record (decimation or uninteresting record).           */
/************************************************************************/
int process_record (int rec_type, int rec_len, char *line)
{
    int status = RECORD_OK;
    int is_mben_rec = 0;
    int is_pben_rec = 0;
    if (verbosity >= 1) fprintf (info, "Enter process_record\n");
    memset ((void *)&u, 0, sizeof(u));

    switch (rec_type) {
    case MBEN_CA_REC:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Record MBEN_CA\n");
        parse_mben (MBEN_CA_REC, (unsigned char *)line, rec_len, &u.mben);
        if (DEBUG(DEBUG_RECORD)) dump_mben (MBEN_CA_REC, "mben", &u.mben, info);
        is_mben_rec = 1;
        break;
    case MBEN_CL_REC:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Record MBEN_CL\n");
        parse_mben (MBEN_CL_REC, (unsigned char *)line, rec_len, &u.mben);
        if (DEBUG(DEBUG_RECORD)) dump_mben (MBEN_CL_REC, "mben", &u.mben, info);
        is_mben_rec = 1;
        break;
    case MBEN_P1_REC:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Record MBEN_P1\n");
        parse_mben (MBEN_P1_REC, (unsigned char *)line, rec_len, &u.mben);
        if (DEBUG(DEBUG_RECORD)) dump_mben (MBEN_P1_REC, "mben", &u.mben, info);
        is_mben_rec = 1;
        break;
    case MBEN_P2_REC:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Record MBEN_P2\n");
        parse_mben (MBEN_P2_REC, (unsigned char *)line, rec_len, &u.mben);
        if (DEBUG(DEBUG_RECORD)) dump_mben (MBEN_P2_REC, "mben", &u.mben, info);
        is_mben_rec = 1;
        break;
    case MBEN_PC_REC:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Record MBEN_PC\n");
        parse_mben (MBEN_PC_REC, (unsigned char *)line, rec_len, &u.mben);
        if (DEBUG(DEBUG_RECORD)) dump_mben (MBEN_PC_REC, "mben", &u.mben, info);
        is_mben_rec = 1;
        break;
    case PBEN_REC:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Record PBEN\n");
        parse_pben (PBEN_REC, (unsigned char *)line, rec_len, &u.pben);
        if (DEBUG(DEBUG_RECORD)) dump_pben (PBEN_REC, "pben", &u.pben, info);
        is_pben_rec = 1;
        break;
    case SNAV_REC:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Record SNAV\n");
        parse_snav (SNAV_REC, (unsigned char *)line, rec_len, &u.snav);
        if (DEBUG(DEBUG_RECORD)) dump_snav (SNAV_REC, "snav", &u.snav, info);
        break;
    case SALM_REC:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Record SALM\n");
        parse_salm (SALM_REC, (unsigned char *)line, rec_len, &u.salm);
        if (DEBUG(DEBUG_RECORD)) dump_salm (SALM_REC, "salm", &u.salm, info);
        break;
    case UNKNOWN_REC:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Unknown record\n");
        status = RECORD_ERROR;
        break;
    default:
        fprintf (info, "Unknown record type %d\n", rec_type);
        status = RECORD_ERROR;
        break;
    }

    /* Perform decimation if requested.                         */
    if (decimate && is_mben_rec && (int)u.mben.seq_tag % (mben_interval) != 0) {
        status = SKIP_RECORD;
    }

    if (decimate && is_pben_rec && (int)u.pben.pbentime % (pben_interval) != 0) {
        status = SKIP_RECORD;
    }

    if (verbosity >= 1) fprintf (info, "Exit process_record rc=0\n");
    return (status);
}

/************************************************************************/
/*  record_string:                                                      */
/*      Return string name for record type, to be used for B2000.       */
/*      String can be specific to record, or generic for receiver.      */
/************************************************************************/
char *record_string (int record_type)
{
    static char *string;
    switch (record_type) {
    case MBEN_CA_REC:
    case MBEN_CL_REC:
    case MBEN_P1_REC:
    case MBEN_P2_REC:
    case MBEN_PC_REC:   string = "MBEN"; break;
    case PBEN_REC:              string = "PBEN"; break;
    case SNAV_REC:              string = "SNAV"; break;
    case SALM_REC:              string = "SALM"; break;
    case RTR_REC:               string = "RTR"; break;
    case UNKNOWN_REC:
    default:            string = "UNKNOWN"; break;
    }
    return (string);
}

/************************************************************************/
/*  end_of_epoch:                                                       */
/*      Return 1 if record signals end of epoch; 0 otherwise.           */
/************************************************************************/
int end_of_epoch (int record_type, void *record) 
{
    return (record_type == PBEN_REC);
}
