/************************************************************************/
/*  Routines to create and encapsulate opaque blockettes (2000).        */
/*  Receiver independent.                                               */
/*                                                                      */
/*      Douglas Neuhauser                                               */
/*      Seismographic Station                                           */
/*      University of California, Berkeley                              */
/*      doug@seismo.berkeley.edu                                        */
/*                                                                      */
/************************************************************************/

/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
/************************************************************************/

#ifndef lint
static char sccsid[] = "@(#)b2000_subs.c        1.1 09/09/97 19:03:43";
#endif

#include <stdio.h>
#include <string.h>
#include <memory.h>
#ifndef _OSK
#include <sys/types.h>
#endif	/* _OSK */
#include <time.h>

#include "gps_generic.h"
#include "gps_seed.h"
#include "b2000_subs.h"
#include "receiver_subs.h"

#ifndef OPAQUE_DATA_FLAGS
#define OPAQUE_DATA_FLAGS       0
#define OPAQUE_RECORD           0x00
#define OPAQUE_STREAM           0x01
#define OPAQUE_PACK             0x00
#define OPAQUE_SINGLE           0x02
#define OPAQUE_NO_SPAN          0x00
#define OPAQUE_FIRST_SPAN       0x04
#define OPAQUE_CONT_SPAN        0x0c
#define OPAQUE_LAST_SPAN        0x08
#define OPAQUE_NO_FILE          0x00
#define OPAQUE_FIRST_FILE       0x10
#define OPAQUE_CONT_FILE        0x20
#define OPAQUE_LAST_FILE        0x30
#endif	/* OPAQUE_DATA_FLAGS */

typedef struct _b2000_fh {
    UWORD       type;                   /*  binary blockette number.    */
    UWORD       next;                   /*  byte offset from sdr to next*/
    UWORD       bl_len;                 /*  total blockette len (bytes).*/
    UWORD       data_offset;            /*  offset to data (bytes).     */
    ULONG       record_num;             /*  record number.              */
    UBYTE       data_word_order;        /*  data word order.            */
    UBYTE       opaque_data_flags;      /*  opaque data flags.          */
    UBYTE       num_hdr_fields;         /*  # of data string hdr fields.*/
}   B2000_FH;
#define B2000_FH_SIZE   15              /*  #of real bytes in fixed hdr.*/

typedef struct _bpr_fh {
    char        location_id[2];
    char        channel_id[3];
    char        flush;
    short int   blockette_len;
    SDR_TIME    btime;
    char        pad[2];
} BPR_FH;

#define FLUSH_RECORD    1
#define NO_FLUSH_RECORD 0
#define MAXLEN_B2000    (512-48-8)      /* max length of B2000.         */

/************************************************************************/
/*  External variables.                                                 */
/************************************************************************/
extern char gps_vendor[];       /* Vendor-specific.                     */
extern char gps_model[];        /* Vendor-specific.                     */
extern char gps_software[];     /* Vendor-specific.                     */
extern char gps_firmware[];     /* Vendor-specific.                     */
extern int verbosity;           /* Verbosity flag.                      */
extern int flush;               /* Flush flag.                          */
extern FILE *info;              /* output file ptr for information.     */

/************************************************************************/
/*  Static variables.                                                   */
/************************************************************************/
static char bpr[MAXLEN_B2000+sizeof(BPR_FH) + sizeof(short int)];
static char b2000[MAXLEN_B2000];
static char vhdr[512];

/************************************************************************/
/*  build_b2000:                                                        */
/*      Create a blockette 2000.                                        */
/*      Assume blockette will fit within maxlen bytes.                  */
/*  Return:                                                             */
/*      number of blockettes as function value.                         */
/*  Set:                                                                */
/*      size of output blockette.                                       */
/************************************************************************/
int build_b2000 
(char           *rec,           /* input opaque data record.            */
 int            rec_len,        /* input opaque data record length.     */
 int            rec_type,       /* vendor-specific record type.         */
 char           *rec_str,       /* vendor-specific record type string.  */
 int            maxlen,         /* max opaque data length for blockette.*/
 char           **outp,         /* ptr to output blockette              */
 int            *outlen)        /* ptr to blockette len (OUT).          */
{
    int hdrlen, vhdrlen;
    B2000_FH *bfh;
    char *p;
    static ULONG record_num = 0;

    if (*outp == NULL) *outp = b2000;
    p = *outp;
    bfh = (B2000_FH *)p;

    if (verbosity >= 2) fprintf (info, "enter b2000\n");
    *outlen = 0;

    /* Check for output buffer overrun for fixed portion of B200 hdr.   */
    if (maxlen < B2000_FH_SIZE) {
        fprintf (stderr, "Error - blockette 2000 fhdr size %d exceeds max size %d\n",
                 B2000_FH_SIZE, maxlen);
        return (0);
    }
    if (record_num > (ULONG)2147483647) record_num = 0;

    /* Fill in fixed portion of B2000 header.                           */
    bfh->type = 2000;   
    bfh->next = 0;
    bfh->bl_len = 15;
    bfh->data_offset = 0;
    bfh->record_num = ++record_num;
    bfh->data_word_order = SEED_BIG_ENDIAN;
    bfh->opaque_data_flags = OPAQUE_RECORD | OPAQUE_PACK | OPAQUE_NO_SPAN | OPAQUE_NO_FILE;
    bfh->num_hdr_fields = 5;
    hdrlen = B2000_FH_SIZE;

    /* Create variable portion of blockette header. */
    if (verbosity >= 3) {
        fprintf (info, "Creating variable header\n");
        fprintf (info, "rec_str='%s'\n", rec_str);    
        fprintf (info, "gps_vendor='%s'\n", gps_vendor);      
        fprintf (info, "gps_model='%s'\n", gps_model);        
        fprintf (info, "gps_software='%s'\n", gps_software);  
        fprintf (info, "gps_firmware='%s'\n", gps_firmware);  
    }
    sprintf (vhdr, "%s~%s~%s~%s~%s~", rec_str, gps_vendor, gps_model, 
             gps_software, gps_firmware);
    if (verbosity >= 3) fprintf (info, "Created variable header\n");
    if ((vhdrlen=strlen(vhdr))+hdrlen > maxlen) {
        fprintf (stderr, "Error - blockette 2000 hdr size %d exceeds max size %d\n",
                 vhdrlen+hdrlen, maxlen);
        return (0);
    }
    memcpy (&p[hdrlen], vhdr, vhdrlen);
    hdrlen += vhdrlen;
    if (verbosity >= 3) fprintf (info, "Created variable header\n");

    /* Add the data to the blockette.                                   */
    bfh->data_offset = hdrlen;
    bfh->bl_len = hdrlen + rec_len;
    if ((int)bfh->bl_len > maxlen) {
        fprintf (stderr, "Error - blockette 2000 size %d exceeds max size %d\n",
                 (int)bfh->bl_len, maxlen);
        return (0);
    }
    if (verbosity >= 3) fprintf (info, "Adding data\n");
    memcpy (&p[hdrlen], (void *)rec, rec_len);

    /* Update returned arguments.                                       */
    *outlen = bfh->bl_len;
    if (verbosity >= 2) fprintf (info, "exit b2000\n");
    return (1);
}

/************************************************************************/
/*  build_bpr:                                                          */
/*      Build a Blockette Piped Record containing opaque data record.   */
/************************************************************************/
int build_bpr
(char           *rec,           /* input opaque data record.            */
 int            rec_len,        /* input opaque data record length.     */
 int            rec_type,       /* vendor-specific record type.         */
 char           *rec_str,       /* vendor-specific record type string.  */
 int            maxlen,         /* max opaque data length for blockette.*/
 char           *location,      /* location for SEED channel.           */
 char           *channel,       /* channel for SEED channel.            */
 SDR_TIME       *btime,         /* ptr to time (possibly NULL).         */
 char           **outp,         /* ptr to output blockette              */
 int            *outlen)        /* ptr to blockette len (OUT).          */
{
    int status;
    int b2000_len;
    short int cksum, *ps, *pe;
    char *b2000 = (char *)&bpr[sizeof(BPR_FH)];
    BPR_FH *bpr_fh = (BPR_FH *)&bpr[0];
    int i, n;
    char *p;
    time_t now;
    struct tm *tm;

    if (verbosity >= 2) fprintf (info, "enter bpr\n");
    memset ((void *)bpr_fh, 0, sizeof(BPR_FH));
    memset (bpr_fh->location_id, ' ', 5);
    if (location) {
        if ((n=strlen(location)) > 2) n = 2;
        strncpy (bpr_fh->location_id, location, n);
    }
    if (channel) {
        if ((n=strlen(channel)) > 3) n = 2;
        strncpy (bpr_fh->channel_id, channel, n);
    }
    switch (flush) {
    case 0:     bpr_fh->flush = NO_FLUSH_RECORD; break;
    case 1:     bpr_fh->flush = FLUSH_RECORD; break;
    default:    bpr_fh->flush = (end_of_epoch(rec_type, rec)) ? 
		    FLUSH_RECORD : NO_FLUSH_RECORD;
    break;
    }
    if (verbosity >= 2) fprintf (info, "Flush: %d\n", bpr_fh->flush);
    *outlen = 0;
    *outp = NULL;
    status = build_b2000 (rec, rec_len, rec_type, rec_str, maxlen,
                          &b2000, &b2000_len);
    if (status != 1) {
        return (0);
    }

    /* Finish BPR_FH, and compute checksum.                             */
    /* Fill record time from current date/time.                         */
    if (btime) {
        memcpy ((void *)&bpr_fh->btime, (void *)btime, sizeof(SDR_TIME));
    }
    else {
        if (verbosity >= 3) fprintf (info, "start time\n");
        now = time(NULL);
#ifndef _OSK
        tm = gmtime(&now);      
#else	/* _OSK */
        tm = localtime(&now);   /* Quanterra sets local time to UTC.    */
#endif	/* _OSK */
        if (verbosity >= 3) fprintf (info, "after time conversion\n");
        bpr_fh->btime.year = tm->tm_year+1900;
        bpr_fh->btime.day  = tm->tm_yday+1;
        bpr_fh->btime.hour = tm->tm_hour;
        bpr_fh->btime.minute = tm->tm_min;
        bpr_fh->btime.second = tm->tm_sec;
        bpr_fh->btime.ticks = 0;
        if (verbosity >= 3) fprintf (info, "end time\n");
    }
    /* Checksum is computed on 16-bit ints on the entire bpr.           */
    /* buffer must be an even number of bytes.                          */
    /* For SEED's sake, we will ensure that it is a multiple of 4,      */
    /* so that subsequent blockettes begin on a 4-byte boundary.        */
    while (b2000_len%4) b2000[b2000_len++] = '\0';      /* round up.    */
    bpr_fh->blockette_len = b2000_len;
    cksum = 0;
    ps = (short int *)&bpr[0];
    pe = (short int *)&b2000[b2000_len];
    while (ps < pe) {
        cksum += *ps++;
    }
    memcpy ((void *)&b2000[b2000_len], (void *)&cksum, sizeof(cksum));
    *outlen = b2000_len + sizeof(BPR_FH) + sizeof(cksum);
    *outp = bpr;
    if (verbosity >= 2) fprintf (info, "exit bpr\n");
    return (1);
}
