/************************************************************************/
/*  Program:    gps_generic                                             */
/*                                                                      */
/*  Description:                                                        */
/*      Read real-time gps binary records and log to a file.            */
/*      Produce optional ASCII dump of gps binary records.              */
/*                                                                      */
/*      Douglas Neuhauser                                               */
/*      Seismographic Station                                           */
/*      University of California, Berkeley                              */
/*      doug@seismo.berkeley.edu                                        */
/*                                                                      */
/************************************************************************/

/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */


/*
Modification History
------------------------------------------------------------------------
95/11/13 DSN    1.0     Initial coding.
96/04/29 DSN    1.1     Modified to create files of specified timespan.
96/06/02 DSN    1.1.1   Improved handling of unknown records.
96/06/19 DSN    1.1.2   Fixed handling of unknown records.
96/06/16 DSN    1.1.3   Add handling of RTR error record.
96/08/30 DSN    1.2     Added decimation capability (specify sample interval).
96/10/01 DSN    1.3     Added copyright msg.  Removed print_syntax function.
96/11/27 DSN    1.4     Decimate PBEN records in addition to MBEN records.
                        Converted to qlib2.
97/08/20 DSN    1.5     Split into generic and vendor-specific code;
                        Ported to Quanterra datalogger.
97/09/09 DSN            Use os/9 create() function to set size of output pipe.
98/10/27 DSN,RB	1.6	Developed trimble version, ported to Quanterra.
*/
/************************************************************************/

#ifndef lint
static char sccsid[] = "%W% %G% %U%";
#endif	/* lint */

#include <stdio.h>

#define VERSION         "1.6 (1998.300)"

#ifndef _OSK
#define DEFAULT_FORMAT_STR      "raw"
#else	/* _OSK */
#define DEFAULT_FORMAT_STR      "bpr"
#define DEFAULT_CONFIG          "/r0/aqcfg"
#endif	/* _OSK */

#ifndef	_OSK
char *syntax[] = {
    "%s version " VERSION " - read and log binary real-time gps data.",
    "%s [-h] [-c config] [-m module] [-n] [-l line] [-b baud] [-p parity]",
    "   [-i n] [-v n] [-d n] [-w n] [-D seconds]",
    "   [-C channel] [-L location] [-f format]",
    "   [-s span] [input_file] [output_file]",
    "where:",
    "   -h          Help - prints syntax message.",
    "   -c config   Specify the name of the config file.",
    "   -m module   Specify the module name in the config file.",
    "               Default module name is the program name.",
    "   -n	    No data output.  Do not open output file or write output",
    "               data even if output file is specified.",
    "   -l line     Specify input serial line.  Mutually exclusive with infile.",
    "   -b baud     Baudrate of input line.  Default is 9600.",
    "   -p parity   Parity of input line:  E, O, or N.  Default is N.",
    "   -i n        Serial line service interval in millisecond increments",
    "               Default = 100",
    "   -v n        Verbosity flag.",
    "   -d n        Debug output (OR of any of the following values:)",
    "                   1 = debug record type",
    "                   2 = debug full record",
    "                   4 = debug input",
    "   -w n        Warn user if no valid data seen in N seconds.",
    "               If n=0, suppress warnings.",
    "   -D seconds  Specify a decimation sample interval in seconds.",
    "               Value must be a multiple of receiver sample interval.",
    "               If option is not specified, save all samples.",
    "   -f format   Specify output format raw, b2000, or bpr.",
    "               Default format is " DEFAULT_FORMAT_STR ".",
    "   -C channel  SEED channel name.",
    "   -L location SEED location name.",
    "   -s span     Timespan of file.  Default is infinity.",
    "               Timespan may positive integer followed by the letter:",
    "               M = minutes, H = hours, d = days",
    "               Minutes must be an even divisor of 1 hour.",
    "   infile      Input file containing ashtech binary data.",
    "               If no input file or input line specified, read from stdin.",
    "               Do not specify input file if line is specified.",
    "   outfile     Output file containing raw output ashtech data,"
    "               or sitename used to construct filename of timespan files.",
    "               If no output file is specified, write data to stdout.",
    NULL };

#else
char *syntax[] = {
    "%s version " VERSION " - read and log binary real-time gps data.",
    "%s [-h] [-c=config] [-m=module] [-n] [-l=line] [-b=baud] [-p=parity]",
    "   [-i=n] [-v=n] [-d=n] [-w=n] [-D=seconds]",
    "   [-C=channel] [-L=location] [-f=format]",
    "   [-s=span] [input_file] [output_file]",
    "where:",
    "   -h          Help - prints syntax message.",
    "   -c=config   Specify the name of the config file.",
    "               Default config file is " DEFAULT_CONFIG ".",
    "   -m=module   Specify the module name in the config file.",
    "               Default module name is the program name.",
    "   -n	    No data output.  Do not open output file or write output",
    "               data even if output file is specified.",
    "   -l=line     Specify input serial line.  Mutually exclusive with infile.",
    "   -b=baud     Baudrate of input line.  Default is 9600.",
    "   -p=parity   Parity of input line:  E, O, or N.  Default is N.",
    "   -i=n        Serial line service interval in millisecond increments",
    "               Default = 100",
    "   -v=n        Verbosity flag.",
    "   -d=n        Debug output (OR of any of the following values:)",
    "                   1 = debug record type",
    "                   2 = debug full record",
    "                   4 = debug input",
    "   -w=n        Warn user if no valid data seen in N seconds.",
    "               If n=0, suppress warnings.",
    "   -D=seconds  Specify a decimation sample interval in seconds.",
    "               Value must be a multiple of receiver sample interval.",
    "               If option is not specified, save all samples.",
    "   -f=format   Specify output format raw, b2000, or bpr.",
    "               Default format is " DEFAULT_FORMAT_STR ".",
    "   -C=channel  SEED channel name.",
    "   -L=location SEED location name.",
    "   infile      Input file containing ashtech binary data.",
    "               If no input file or input line specified, read from stdin.",
    "               Do not specify input file if line is specified.",
    "   outfile     Output file containing raw output ashtech data,"
    "               or sitename used to construct filename of timespan files.",
    "               If no output file is specified, write data to stdout.",
    NULL };
#endif	/* _OSK */

#include <stdlib.h>
#include <string.h>
#include <memory.h>
#include <signal.h>
#include <errno.h>
 
#ifndef _OSK
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#else	/* _OSK */
#include <types.h>
#include <sysio.h>
#include <modes.h>
/* time.h included by stdlib.h */
#endif	/* _OSK */

#ifndef _OSK
#include <qlib2.h>
#endif	/* _OSK */

#include <dpstruc.h>
#include <cfgutil.h>

#include "gps_generic.h"
#include "gps_seed.h"
#include "b2000_subs.h"
#include "receiver_subs.h"
#include "serial_io.h"

#ifndef _OSK
#define TIMESTAMP(out,time)     fprintf(stderr, "%s ", time_to_str(time,date_fmt))
FILE *out = NULL;               /* output file ptr.                     */
#else	/* _OSK */
int out = -10;                  /* output file descriptor.              */
#define TIMESTAMP(out,time)
#define	SIGUSER1		260
#define	ALARMTIME(secs)		((secs << 8) | 0x80000000)
#define PIPE_SIZE               8192        /* size of output pipe.     */
/* Some missing functions in the C string library. */
#define strrchr rindex
#define strchr  index
int strcasecmp (const char *, const char *);
int strncasecmp (const char *, const char *, int);
#endif	/* _OSK */

#define SKIPEQ  if (*optarg=='=')++optarg

#define RAW_FORMAT              1
#define BLOCKETTE_FORMAT        2
#define QUANTERRA_BPR_FORMAT    3

/************************************************************************/
/*  External variables defined elsewhere.                               */
/************************************************************************/
extern char input_rec[];        /* line to hold input binary records.   */
extern char gps_vendor[];       /* Vendor name.                         */
extern char gps_model[];        /* model name.                          */
extern char gps_software[];     /* software name.                       */
extern char gps_firmware[];     /* firmware name.                       */
extern char out_pattern[];      /* outfile pattern template.            */
extern int maxrlen;             /* max input record length.             */

/************************************************************************/
/*  External variables.                                                 */
/************************************************************************/
char *cmdname;                  /* program name.                        */
char configfile[CFGWIDTH];      /* config file.                         */
char module[CFGWIDTH];          /* module/section name.                 */
char sitename[CFGWIDTH];        /* output file site for spanned file.   */
char port[CFGWIDTH];            /* input serial device name.            */
char infile[CFGWIDTH];          /* input file name.                     */
char outfile[CFGWIDTH];         /* output file name.                    */
char span[CFGWIDTH];            /* timespan for file.                   */
FILE *in = stdin;               /* input file descriptor;               */
FILE *info = stdout;            /* output file ptr for information.     */
int no_output = 0;		/* flag to suppress data output.	*/
int infd = 0;                   /* input file descriptor.               */
int flush = 0;                  /* flush flag.                          */
int finish_proc = 0;            /* flag for signal handler.             */
int debug_option = -1;          /* default debug option.                */
int verbosity = -1;             /* verbosity flag.                      */
int warn = -1;			/* warning time interval for no data.	*/
int warn_now = 0;		/* issue a warning statement now.	*/
int warned = 0;			/* true if we have issued warning.	*/
time_t time_last_record;	/* time of last good record (or start).	*/
char *__progname;               /* program name (for BSD getopt).       */
#ifndef _OSK
INT_TIME span_begtime;          /* beginning time of timespan.          */
INT_TIME span_endtime;          /* end time of timespan.                */
int date_fmt = JULIAN_FMT_1;    /* format for date/time string.         */
#else	/* _OSK */
int alarmid = -1;		/* alarmid for OS9 alarms.		*/
#endif	/* _OSK */

#ifndef _OSK
/************************************************************************/
/*  qnow_gm:                                                            */
/*      Return current gm time in INT_TIME structure.                   */
/************************************************************************/
INT_TIME qnow_gm()
{
    time_t clock;
    time (&clock);
  
    return (int_time_from_time_tm(gmtime(&clock)));
}

/************************************************************************/
/*  alarm_handler:							*/
/*      Signal handler to print alarm for no data.			*/
/************************************************************************/
void alarm_handler(int sig)
{
    time_t now;
    now = time(NULL);
    if (warned < 10) {
	TIMESTAMP(stderr,qnow_gm());
	fprintf (stderr, "Warning: No valid GPS record for %d seconds.\n",
		 now - time_last_record);
	++warned;
	if (warned == 10) {
	    TIMESTAMP(stderr,qnow_gm());
	    fprintf (stderr, "Warning: Additional missing GPS record messages will be suppressed.\n");
	}
	fflush (stderr);
	alarm (warn);
    }
    signal (sig,alarm_handler);    /* Re-install handler (for SVR4)    	*/
}

/************************************************************************/
/*  finish_handler:                                                     */
/*      Signal handler to finish reading record and exit.               */
/************************************************************************/
void finish_handler(int sig)
{
    finish_proc = 1;
    signal (sig,finish_handler);    /* Re-install handler (for SVR4)	*/
}

#else	/* _OSK */
/************************************************************************/
/*  sighand:								*/
/*  Signal handler for OS/9.						*/
/*	Note that C I/O is not allowed in the signal handler if it is	*/
/*	used elsewhere in the program.					*/
/************************************************************************/
int sighand(int sigcode) 
{
    time_t now;
    switch (sigcode) {
    case SIGWAKE:	
    case SIGHUP:	return;
    case SIGQUIT:
    case SIGINT:	exit(sigcode);
    case SIGUSER1:
	alarmid = -1;
	if (warned <= 10) {
	    ++warned;
	    /* Set warn_now to the time since last good record. */
	    warn_now = 1;
	    if ((alarmid = alm_set (SIGUSER1,ALARMTIME(warn))) == -1) {
		exit(50);
	    }
	}
	return (0);
	break;
    default:    exit(1);
    }
}
#endif	/* _OSK */

#ifndef _OSK
/************************************************************************/
/*  validate_span:                                                      */
/*      Verify that span is valid.                                      */
/*      Set start and end time of span.                                 */
/************************************************************************/
int validate_span (char *duration)
{
    int i;
    char *p;
    int error = 0;

    i = strtol(duration,&p,10);
    if (strcmp(p,"H") == 0) {
	if (i <= 0 || i > 24 || 24%i != 0) ++error;
    }
    else if (strcmp(p,"M") == 0) {
	if (i <= 0 || i > 60 || 60%i != 0) ++error;
    }
    else if (strcmp(p,"d") == 0) {
	if (i <= 0) ++error;
    }
    else ++error;
  
    return (error);
}

/************************************************************************/
/*  qnow_gmt:                                                           */
/*      Return current GMT time in INT_TIME format.                     */
/************************************************************************/
INT_TIME qnow_gmt ()
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (int_time_from_timeval(&tv));
}

/************************************************************************/
/*  set_span_bounds:                                                    */
/*      Compute boundaries of timespan.                                 */
/*      Set global variables spanbeg, spanend;                          */
/************************************************************************/
void set_span_bounds (char *duration)
{
    EXT_TIME et;
    INT_TIME qnow;
    int i;
    i = 0;
    qnow = qnow_gmt();
    et = int_to_ext(qnow);
    et.hour = et.minute = et.second = et.usec = 0;
    span_begtime = ext_to_int(et);
    span_endtime = end_of_span(span_begtime, duration);
    while (tdiff(qnow, span_endtime) >= 0.) {
	span_begtime = span_endtime;
	span_endtime = end_of_span (span_begtime, duration);
    }
}

/************************************************************************/
/*  build_filename:                                                     */
/*      Build a filename from the template format.                      */
/************************************************************************/
char *build_filename(char *site, char *pattern, INT_TIME btime)
{
    static char name[1024];
    char *ip, *op;
    EXT_TIME begtime;
    int i;

    begtime = int_to_ext(btime);
    memset ((void *)name, 0, sizeof(name));
    ip = pattern;
    op = name;

    while (*ip != '\0') {
	switch (*ip) {
	case ('%'):
	    switch (*++ip) {
	    case ('S'):
		sprintf (op, "%-4.4s", site);
		uppercase(op);
		/* Convert blanks to '_' (for filename) */
		for (i=0; i<4; i++) {
		    if (op[i] == ' ') op[i] = '_';
		}
		op += 4; break;
	    case ('s'):
		sprintf (op, "%-4.4s", site);
		lowercase(op);
		/* Convert blanks to '_' (for filename) */
		for (i=0; i<4; i++) {
		    if (op[i] == ' ') op[i] = '_';
		}
		op += 4; break;
	    case ('Y'):
		sprintf (op, "%04d", begtime.year);
		op += 4; break;
	    case ('y'):
		sprintf (op, "%02d", begtime.year % 100);
		op += 2; break;
	    case ('m'):
		sprintf (op, "%02d", begtime.month);
		op += 2; break;
	    case ('d'):
		sprintf (op, "%02d", begtime.day);
		op += 2; break;
	    case ('j'):
		sprintf (op, "%03d", begtime.doy);
		op += 3; break;
	    case ('H'):
		sprintf (op, "%02d", begtime.hour);
		op += 2; break;
	    case ('M'):
		sprintf (op, "%02d", begtime.minute);
		op += 2; break;
	    default:
		fprintf (info, "invalid format\n");
		sprintf (op, "%%%c", *ip);
		break;
	    }
	    ++ip; break;
	default:    *op++ = *ip++; break;
	}
    }
    *op = '\0';
    return(name);
}

/************************************************************************/
/*  within_timespan:                                                    */
/*      Check whether we are within the current timespan.               */
/*      Return 1 if within timespan, 0 if outside timespan.             */
/************************************************************************/
int within_timespan ()
{
    INT_TIME qnow;
    if (span == NULL || span[0] == '\0' ) return (1);
    /* Check to see if we need to switch files.                         */
    qnow = qnow_gmt();
    return (tdiff(qnow,span_endtime)<=0);
}

/************************************************************************/
/*  open_output:                                                        */
/*      Open output file if necessary.                                  */
/*      Sets global variables: outfile                                  */
/************************************************************************/
FILE *open_output (FILE *out, char *sitename, char *pattern, char *duration)
{
    if (no_output) return (NULL);
    if (sitename == NULL || sitename[0] == '\0') return (NULL);
    if (duration == NULL || duration[0] == '\0') {
	/* Single output file.                                          */
	strcpy(outfile, sitename);
    }
    else {
	/* Output file dependent on timespan.                           */
	if (! within_timespan()) {
	    set_span_bounds (duration);
	    if (out) fclose(out);
	    out = NULL;
	    outfile[0] = '\0';
	}
	if (out == NULL) {
	    strcpy (outfile, build_filename (sitename, pattern, span_begtime));
	}
    }
    if (out == NULL) out = fopen(outfile,"a");
    if (out == NULL) {
	fprintf (stderr, "Unable to open output file %s\n", outfile);
	exit(1);
    }
    if (verbosity >= 1) fprintf (info, "Opened output file %s\n", outfile);
    return (out);
}
#else	/* _OSK */
/************************************************************************/
/*  open_output:                                                        */
/*      Open output file if necessary.                                  */
/*      Sets global variables: outfile                                  */
/************************************************************************/
int open_output (int out, char *sitename, char *pattern, char *duration)
{
    if (no_output) return (NULL);
    if (sitename == NULL || sitename[0] == '\0') return (-1);
    /* Single output file.                                              */
    strcpy (outfile, sitename);
    if (out < 0) {
	if (access(sitename, 0) >= 0)
	    out = open (outfile, (short int)S_IWRITE);
	else
	    out = create (outfile, (short int)S_IWRITE|S_ISIZE,
			  S_IREAD|S_IWRITE, PIPE_SIZE);
    }
    if (out < 0) {
	fprintf (stderr, "Unable to open output file %s\n", outfile);
	exit(1);
    }
    if (verbosity >= 1) fprintf (stderr, "Opened output file %s\n", outfile);
    return (out);
}
#endif	/* _OSK */

/************************************************************************/
/*  main routine.                                                       */
/************************************************************************/
int main (int argc, char **argv)
{
    int rec_len, nskip;
    int rec_type;
    char *rec_string;
    int status, ok;
    char *p;
    int baud = 0;
    char parity = '\0';
    int flag;
    int service_time = -1;             /* input service time in msec.  */
    int sample_interval = 0;
    int decimate = 0;
    char *output_rec;
    int output_len;
    int format = 0;
    char format_str[CFGWIDTH] = "";
    char channel[CFGWIDTH] = "";
    char location[CFGWIDTH] = "";
    config_struc cs;
    char str1[CFGWIDTH], str2[CFGWIDTH];
    time_t now;
  
    /* Variables needed for getopt. */
    extern char *optarg;
    extern int  optind, opterr;
    int         c;
  
    infd = fileno(stdin);
    cmdname = ((p = strrchr(*argv,'/')) != NULL) ? ++p : *argv;
    __progname = cmdname;
#ifdef	_OSK
    strcpy(configfile,DEFAULT_CONFIG);
#endif	/* _OSK */
    while ( (c = getopt(argc,argv,"hc:m:nf:d:v:w:s:C:L:l:b:p:i:D:")) != -1)
	switch (c) {
	case '?':
	case 'h': print_syntax(cmdname,syntax,info); exit(0); break;
	case 'c': SKIPEQ; strcpy(configfile,optarg); break;
	case 'm': SKIPEQ; strcpy(module,optarg); break;
	case 'n': no_output = 1; break;
	case 'f': SKIPEQ; strcpy(format_str,optarg); break;     
	case 'd': SKIPEQ; debug_option = atoi(optarg); break;
	case 'v': SKIPEQ; verbosity = atoi(optarg); break;
	case 'w': SKIPEQ; warn = atoi(optarg); break;
	case 'C': SKIPEQ; strcpy(channel,optarg); break;
	case 'L': SKIPEQ; strcpy(location,optarg); break;
	    /* Serial line attributes   */
	case 'l': SKIPEQ; strcpy(port,optarg); break;
	case 'b': SKIPEQ; baud = atoi(optarg); break;
	case 'p': SKIPEQ; parity = optarg[0]; break;
	case 'i': SKIPEQ; service_time = atoi(optarg); break;
	case 'D': SKIPEQ; sample_interval = atoi(optarg); decimate=1; break;
#ifndef _OSK
	case 's': SKIPEQ; strcpy(span,optarg); break;
#endif	/* _OSK */
	default:  fprintf (stderr, "Unknown option: -%c\n", c); exit(1);
	}

    /*  Skip over all options and their arguments.                      */
    argv = &(argv[optind]);
    argc -= optind;

    if (module[0] == '\0') strcpy (module,cmdname);
    TIMESTAMP(stderr,qnow_gm());
    fprintf (stderr, "%s version %s startup\n", module, VERSION);

    if (strlen(configfile)) {
	if (open_cfg (&cs, configfile, module)) {
	    fprintf (stderr, "Unable to open config file %s\n", configfile);
	    exit(1);
	}
	/* Allow command line options to override values when appropriate. */
	while (1) {
	    read_cfg(&cs, str1, str2);
	    if (str1[0] == '\0') break;
	    else if (strcasecmp(str1,"INPUT") == 0) {
		strcpy(infile,str2);
	    }
	    else if (strcasecmp(str1,"PORT") == 0) {
		if (port[0] == '\0') strcpy(port,str2);
	    }
	    else if (strcasecmp(str1,"BAUD") == 0) {
		if (baud == 0) baud = atoi(str2);      
	    }
	    else if (strcasecmp(str1,"PARITY") == 0) {
		if (parity == '\0') parity = str2[0];
	    }
	    else if (strcasecmp(str1,"OUTPUT") == 0) {
		strcpy (sitename,str2);
	    }
	    else if (strcasecmp(str1,"WARNAT") == 0) {
		if (warn < 0) warn = atoi(str2);
	    }
	    else if (strcasecmp(str1,"debug") == 0) {
		if (debug_option < 0) debug_option = atoi(str2);
	    }
	    else if (strcasecmp(str1,"FORMAT") == 0) {
		if (format_str[0] == '\0') strcpy(format_str,str2);
	    }
	    else if (strcasecmp(str1,"CHANNEL") == 0) {
		if (channel[0] == '\0') strcpy(channel,str2);
	    }
	    else if (strcasecmp(str1,"LOCATION") == 0) {
		if (location[0] == '\0') strcpy(location,str2);
	    }
	    else if (strcasecmp(str1,"VENDOR") == 0) {
		strcpy(gps_vendor,str2);
	    }
	    else if (strcasecmp(str1,"MODEL") == 0) {
		strcpy(gps_model,str2);
	    }
	    else if (strcasecmp(str1,"SOFTWARE") == 0) {
		strcpy(gps_software,str2);
	    }
	    else if (strcasecmp(str1,"FIRMWARE") == 0) {
		strcpy(gps_firmware,str2);
	    }
	    else if (strcasecmp(str1,"FLUSH") == 0) {
		if (str2[0]=='n' || str2[0]=='N') flush = 0;
		else if (str2[0]=='y' || str2[0]=='Y') flush = 1;
		else if (str2[0]=='e' || str2[0]=='E') flush = 2;
		else flush = 0;
	    }
	    else if (strcasecmp(str1,"SERVICETIME") == 0) {
		if (service_time < 0) service_time = atoi(str2);
	    }
	    else if (strcasecmp(str1,"DECIMATE") == 0) {
		if (sample_interval < 0) {
		    sample_interval = atoi(str2); 
		    decimate = 1;
		}
	    }
	    else if (strcasecmp(str1,"VERBOSITY") == 0) {
	    if (verbosity < 0) verbosity = atoi(str2);
	    }
#ifndef _OSK
	    else if (strcasecmp(str1,"SPAN") == 0) {
		strcpy (span,str2);
	    }
#endif	/* _OSK */
	    else fprintf (stderr, "Unknown directive: %s\n", str1);
	}
	close_cfg(&cs);
    }
  
    /* Establish defaults if none provided. */
    if (debug_option < 0) debug_option = 0;
    if (verbosity < 0) verbosity = 0;
    if (warn < 0) warn = 0;
    if (baud == 0) baud = 9600;
    if (parity == '\0') parity = 'N';
    if (format_str[0] == '\0') strcpy (format_str,DEFAULT_FORMAT_STR);

    if (service_time < 0) service_time = 100; /* input service time in msec. */

    if (strcasecmp(format_str,"raw")==0) format = RAW_FORMAT;
    else if (strcasecmp(format_str,"b2000")==0) format = BLOCKETTE_FORMAT;
    else if (strcasecmp(format_str,"bpr")==0) format = QUANTERRA_BPR_FORMAT;
    if (format == 0) {
	fprintf (stderr, "Invalid or missing format\n");
	exit(1);
    }
    if (verbosity >= 1) fprintf (info, "format=%d\n", format);

    if (setup_service_time (service_time) < 0) {
	fprintf (stderr, "Error - service interval must be >= 1\n");
	exit(1);
    }
    
    if (setup_decimation (decimate, sample_interval) != 0) {
	exit(1);
    }

#ifndef _OSK
    /* Process time span for output file.                               */
    if (span && span[0]) {
	if (validate_span(span) != 0) {
	    fprintf (stderr, "Invalid span: %s\n", span);
	    exit(1);
	}
	set_span_bounds (span);
    }
#endif	/* _OSK */

    /* Process input file or input serial line.                         */
    /* If port was not specified, the next argument is input file.	*/
    /* If no port or input file specified, input is read from stdin.	*/

    if (strlen(port)) {
	infd = open_serial (port, baud, parity);
    }
    else {
	if (argc-- > 0) strcpy(infile, *argv++);
	if (strlen(infile) > 0) {
	    in = fopen (infile, "r");
	    if (DEBUG(DEBUG_INPUT) || verbosity > 0) {
		fprintf (info, "Opened file %s\n", infile);
	    }
	    infd = fileno(in);
	}
    }
    /* Check to ensure that we have an input.                           */
    if (infd < 0) {
	fprintf (stderr, "Unable to open infile %s\n", infile);
	exit(1);
    }

    /* Process output file.                                             */
    /* If argument specified, it is either full output filename or	*/
    /* sitename used in construction of time spanned output filename.	*/
    /* Argument is interpreted as sitename if -s option was specified.	*/
    /* If no output is specified, output file is stdout.		*/

    if (argc-- > 0) strcpy(sitename,*argv++);
    if (sitename[0]=='\0' && span[0]) {
	fprintf (stderr, "Error - missing output file for span\n");
	exit(1);
    }

    if (sitename[0]=='\0' && ! no_output) {    
	info = stderr;
#ifndef _OSK
	out = stdout;
#else	/* _OSK */
	out = 1;
#endif	/* _OSK */
    }

    time_last_record = time(NULL);
#ifndef _OSK
    signal (SIGHUP,finish_handler);
    signal (SIGALRM,alarm_handler);
    if (warn) alarm(warn);
#else	/* _OSK */
    intercept(sighand);
    if (warn) {
  	if ((alarmid = alm_set (SIGUSER1,ALARMTIME(warn))) == -1) {
	    fprintf (stderr, "Error setting alarm\n");
	    exit(1);
  	}
    }
#endif	/* _OSK */

    while ((status=get_bin_record(input_rec, maxrlen, &rec_type, &rec_len, &nskip, infd)) != EOF
	   && (! finish_proc)) {

	if (nskip > 0) {
	    TIMESTAMP(stderr,qnow_gm());
	    fprintf (stderr, "Skipped %d bytes\n", nskip);
	    fflush (stderr);
	}
	
	/* Handle errors returned from reading record.                  */
	if (status < 0) {
	    TIMESTAMP(stderr,qnow_gm());
	    switch (status) {
	    case RECORD_ERROR:
		input_rec[rec_len] = '\0';
		fprintf (stderr, "Record error: %s\n", input_rec);
		break;
	    case CKSUM_ERROR:
		fprintf (stderr, "Checksum error\n");
		break;
	    case TERM_ERROR:
		fprintf (stderr, "Input record error - missing record terminator\n");
		break;
	    default:
		fprintf (stderr, "Unknown error: %d\n", status);
		break;
	    }
	    continue;
	}
	
	flag = process_record (rec_type, rec_len, input_rec);
	if (flag < 0) continue;
	if (flag & SKIP_RECORD) continue;
	/* Output binary record if requested by cmdline options.        */    
	if (sitename[0] && ! no_output) {
#ifndef _OSK
	    if ((out = open_output (out, sitename, out_pattern, span)) == NULL) {
		TIMESTAMP(stderr,qnow_gm());
		fprintf (stderr, "Error %d opening output file %s\n", errno, outfile);
		exit(1);
	    }
#else	/* _OSK */
	    if ((out = open_output (out, sitename, out_pattern, span)) < 0) {
		TIMESTAMP(stderr,qnow_gm());
		fprintf (stderr, "Error %d opening output file %s\n", errno, outfile);
		exit(1);
	    }
#endif	/* _OSK */
	}

	/* Disable alarm. */
#ifndef	_OSK
	if (warn) alarm (0);
#else	/* _OSK */
	if (warn && alarmid != -1) { 
	    if (alm_delete(alarmid) == -1) {
		fprintf (stderr, "Error disabling alarm, errno=%d\n", 
			 errno);
		exit(1);
	    }
	}
#endif	/* _OSK */

	now = time(NULL);
	if (warned) {
	    TIMESTAMP(stderr,qnow_gm());
	    fprintf (stderr, "GPS record received after %d seconds.\n",
		     now - time_last_record);
	}
	time_last_record = now;
	warned = 0;
	warn_now = 0;

	/* Re-enable alarm. */
#ifndef	_OSK
	if (warn) alarm (warn);
#else	/* _OSK */
	if (warn) {
	    if ((alarmid = alm_set (SIGUSER1,ALARMTIME(warn))) == -1) {
  		fprintf (stderr, "Error setting alarm\n");
  		exit(1);
	    }
	}
#endif	/* _OSK */

	/* Create output record from input record.                      */
	switch (format) {
	case RAW_FORMAT:        
	    output_rec = input_rec;
	    output_len = rec_len;
	    ok = 1;
	    break;
	case BLOCKETTE_FORMAT:
	    output_rec = NULL;
	    rec_string = record_string (rec_type);
	    ok = build_b2000 (input_rec, rec_len, rec_type, rec_string,
			      512-65, &output_rec, &output_len);
	    break;
	case QUANTERRA_BPR_FORMAT:
	    rec_string = record_string (rec_type);
	    ok = build_bpr (input_rec, rec_len, rec_type, rec_string,
			    512-65, location, channel, NULL, 
			    &output_rec, &output_len);
	    break;
	default:
	    ok = 0;
	    break;
	}
	if (! ok) {
	    TIMESTAMP(stderr,qnow_gm());
	    fprintf (stderr, "Unable to create output record\n");
	    continue;
	}
	
	/* Write output record.                                 */
	if (no_output) continue;
#ifndef _OSK
	if (fwrite ((void *)output_rec, output_len, 1, out) != 1) {
	    TIMESTAMP(stderr,qnow_gm());
	    fprintf (stderr, "Error writing record to %s\n", outfile);
	    exit(1);
	}
	fflush (out);
#else	/* _OSK */
	if (write (out, (void *)output_rec, output_len) != output_len) {
	    TIMESTAMP(stderr,qnow_gm());
	    fprintf (stderr, "Error writing record to %s\n", outfile);
	    exit(1);
	}
#endif	/* _OSK */
    }

    /* Determine reason for exiting loop, and set program return status.*/
    if (finish_proc) {
	status = 0;
    }
    else if (status == EOF) {
	if (DEBUG(DEBUG_INPUT)) fprintf (info, "EOF reached\n");
	status = 0;
    }
    else {
	fprintf (info, "Error status %d\n", status);
    }
    return (status);
}

#ifdef  _OSK
/************************************************************************/
/*  print_syntax:                                                       */
/*      Print the syntax description of program.                        */
/************************************************************************/
int print_syntax
(char        *cmd,           /* program name.                        */
 char        *syntax[],      /* syntax array.                        */
 FILE        *fp)            /* FILE ptr for output.                 */
{
    int i;
    for (i=0; syntax[i] != NULL; i++) {
	fprintf (fp, syntax[i], cmd);
	fprintf (fp, "\n");
    }
    return (0);
}
#endif	/* _OSK */
