/************************************************************************/
/*  Routines for serial i/o devices.                                    */
/*  Receiver independent.                                               */
/*                                                                      */
/*      Douglas Neuhauser                                               */
/*      Seismographic Station                                           */
/*      University of California, Berkeley                              */
/*      doug@seismo.berkeley.edu                                        */
/*                                                                      */
/************************************************************************/

/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/*
Modification History
------------------------------------------------------------------------
97/09/09 DSN            Use Quanterra blockread() to read from serial port.
*/
/************************************************************************/

#ifndef lint
static char sccsid[] = "%W% %G% %U%";
#endif	/* lint */

#include <stdio.h>
#include <errno.h>
#include <ctype.h>

#ifndef _OSK
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <qlib2.h>
#ifdef  TERMIO
#include <termio.h>
#else	/* TERMIO */
#include <termios.h>
#endif	/* TERMIO */
#else	/* _OSK */
/* ststat.h included by stdlib.h */
#include <sysio.h>
#include <modes.h>
#include <sgstat.h>
#endif	/* _OSK */
#ifndef _OSK
#include <sys/types.h>
#endif	/* _OSK */
#include <time.h>

#include "gps_generic.h"
#include "receiver_subs.h"
#include "serial_io.h"

#ifndef _OSK
#define TIMESTAMP(out,time)     fprintf(stderr, "%s ", time_to_str(time,date_fmt))
#ifndef USLEEPTIME
#define USLEEPTIME  100000
#endif	/* USLEEPTIME */
#else	/* _OSK */
#define TIMESTAMP(out,time)
#define ERROR -1
#endif	/* _OSK */

/************************************************************************/
/*  Externals                                                           */
/************************************************************************/
extern FILE *info;              /* output file ptr for information.     */
extern debug_option;            /* default debug option.                */
extern int warn;		/* warning time interval for no data.	*/
extern int warn_now;		/* issue a warning statement now.	*/
extern int warned;			/* true if we have issued warning.	*/
extern time_t time_last_record;	/* time of last good record (or start).	*/
#ifndef _OSK
extern int date_fmt;            /* format for date/time string.         */
#endif	/* _OSK */

/************************************************************************/
/*  Static variables.                                                   */
/************************************************************************/
#define BLOB        8192
static unsigned char sbuf[BLOB];
static unsigned char *src = sbuf;
static unsigned char *srcend = sbuf;
static int serial_io = 0;       /* input comes from serial io.          */
#ifndef _OSK
int unix_service_time;          /* async service time in .1 seconds     */
#else	/* _OSK */
unsigned int pollslp;           /* async service time in 1/256 sec.     */
#endif	/* _OSK */

/************************************************************************/
/*  External function declarations.                                     */
/************************************************************************/
#ifndef _OSK
INT_TIME qnow_gm();
#else	/* _OSK */
void qnow_gm();
#endif	/* _OSK */

/************************************************************************/
/*  Internal function declarations.                                     */
/************************************************************************/
static int fillbuf (int fd);
static int nextchar (int fd, unsigned char *p);

/************************************************************************/
/*  setup_service_time:                                                 */
/*      Establish polling service time on serial port.                  */
/************************************************************************/
int setup_service_time (int polltime)
{
    if (polltime <= 0) return (-1);
#ifndef _OSK
    unix_service_time = polltime / 100;
    if (unix_service_time < 1) unix_service_time = 1;
#else	/* _OSK */
    /* OS-9 used 1/256 second increments. */
    polltime = polltime * 1000;     /* convert to microsecends.         */
    /* If high order bit is 1, it indicates number is # of 1/256 seconds. */
    /* Value of 1 indicates release timeslice. */
    if (polltime < 11000) 
        pollslp = 1;    /* 1 timeslice */
    else 
        pollslp = 0x80000000 | (polltime / 3906);       
#endif	/* _OSK */
    return (0);
}

#ifdef  _OSK
/************************************************************************/
/*  fillbuf:                                                            */
/*      Fill input buffer from input device.                            */
/************************************************************************/
static int fillbuf (int fd)
{
    int numread, count;
    int err;
    time_t now;
    
    src = sbuf;
    srcend = src;
    count = BLOB;

    /* OS/9 signal handlers cannot perform C I/O if it is used		*/
    /* elsewhere in the code, so perform warning here.			*/
    if (warn_now) {
	now = time(NULL);
	if (warned <= 10) {
	    fprintf (stderr, "Warning: No valid GPS data for %d seconds.\n",
		     now - time_last_record);
	}
	if (warned == 10) {
	    fprintf (stderr, "Warning: Additional missing GPS record messages will be suppressed.\n");
	}
	warn_now = 0;
    }

    /* For OS/9 serial devices, query the number of available char.     */
    /* Read only the available number of characters.                    */
    if (serial_io) {
        count = _gs_rdy(fd);
        if (DEBUG(DEBUG_INPUT)) {
            fprintf (info, "_gs_rdy=%d\n", count);
        }
        if (count <= 0) {
            if (DEBUG(DEBUG_INPUT)) {
                fprintf (info, "no read\n");
            }
            return (0);
        }
/*:: _gs_rdy cannot distinguish EOF or invalid fd from no characters.
	if (count < 0) {
	    if (DEBUG(DEBUG_INPUT)) {
		fprintf (info, "EOF on serial line fd=%d: count=%d\n", fd, count);
	    }
	    return (EOF);
	}
::*/
    }
    if (count > BLOB) count = BLOB;
    /* blockread() is available only for Quanterra serial drivers,      */
    /* but is more efficient than read(). Note reversed arg order.      */
#ifdef	BLOCKREAD
    err = blockread (fd, count, src) ;
    if (err == 208) 
#else	/* BLOCKREAD */
	err = 0;
    if (1)
#endif	/* BLOCKREAD */
        numread = read(fd, src, count);
    else if (err == 0)
        numread = count ;
    else {
        numread = -1 ;
        errno = err ;
    }
    if (DEBUG(DEBUG_INPUT)) {
        fprintf (info, "%d bytes read\n", numread);
    }
    if (numread > 0) {
        srcend = srcend + numread;
        return (0);
    }
    else if (numread < 0) {
        if (DEBUG(DEBUG_INPUT)) {
            fprintf (info, "expected %d char, got %d\n",
                     count, numread);
        }
        return (0);
    }
    else {
        return ((serial_io) ? 0 : EOF);
    }
}
#else	/* _OSK */
/************************************************************************/
/*  fillbuf:                                                            */
/*      Fill input buffer from input device.                            */
/************************************************************************/
static int fillbuf (int fd)
{
    int numread, count;
    int err;
    
    src = sbuf;
    srcend = src;
    count = BLOB;

    numread = read(fd, src, count);
    if (DEBUG(DEBUG_INPUT)) {
        fprintf (info, "%d bytes read\n", numread);
    }
    if (numread > 0) {
        srcend = srcend + numread;
        return (0);
    }
    else if (numread < 0 && errno != EINTR) {
        if (errno != EWOULDBLOCK)
            close (fd);
        perror ("Error reading from input ");
        return (-1);
    }
    else {
        return ((serial_io) ? 0 : EOF);
    }
}
#endif	/* _OSK */

/************************************************************************/
/*  nextchar:                                                           */
/*      Return next input character in location specified by ptr.       */
/*      Return 1 if character available, 0 of none, -1 on EOF.          */
/************************************************************************/
static int nextchar (int fd, unsigned char *p)
{
    int status;
    if (src == srcend) {
        status = fillbuf(fd);
        if (status < 0) return (EOF);
    }
    if (src > srcend) {
        fprintf (stderr, "Error: src > srcend\n");
        exit(1);
    }
    if (src != srcend) {
        *p = *src++;
        return (1);
    }
    else {
        return 0;
    }
}

/************************************************************************/
/*  nextc:                                                              */
/*      Function that operates similar to getc on file descriptor.      */
/*      Returns next character or EOF.                                  */
/************************************************************************/
int nextc (int fd)
{
    int status;
    unsigned char c;
    unsigned int ic;
    int counter = 0;
    time_t now;
    int losstime;
    while (1) {
        status = nextchar(fd, &c);
        if (status == 1) {
            if (DEBUG(DEBUG_INPUT)) {
                ic = c;
                if (! isprint(ic)) ic = ' ';
                fprintf (info, "c = %c (%02x)\n", ic, c);
            }
            return (c);
        }
        if (status == -1) return (EOF);
#ifdef  _OSK
        tsleep (pollslp);
#else	/* _OSK */
#ifdef  POLL
        usleep(USLEEPTIME*unix_service_time);
#endif	/* POLL */
#endif	/* _OSK */
    }
}

#ifndef	_OSK
/************************************************************************/
/*  open_serial -                                                       */
/*      Open a serial line for reading.                                 */
/************************************************************************/
int open_serial (char *line, int baud, char parity)
{
    int fd;
    int mflags, bis;
    int status = 0;
#ifdef  TERMIO
    struct termio sttynew;
    struct termio sttyold;
#else	/* TERMIO */
    struct termios sttynew;
    struct termios sttyold;
#endif	/* TERMIO */

    unsigned int cflag = 0;

    if ((fd = open (line, O_RDONLY, 0666)) < 0) {
        TIMESTAMP(stderr,qnow_gm());
        fprintf (stderr, "Error opening line %s\n", line);
        return (-1);
    }
    if (DEBUG(DEBUG_INPUT)) {
        fprintf (info, "Opened serial line %s fd=%d\n", line, fd);
    }
#ifdef  TERMIO
    if (ioctl (fd, TCGETA, &sttyold) < 0) {
        TIMESTAMP(stderr,qnow_gm());
        fprintf (stderr, "Error getting settings for line %s\n", line);
        close (fd);
        return (-1);
    }
#else	/* TERMIO */
    if (tcgetattr(fd, &sttyold) < 0) {
        TIMESTAMP(stderr,qnow_gm());
        fprintf (stderr, "Error getting settings for line %s\n", line);
        close (fd);
        return (-1);
    }
#endif	/* TERMIO */
    memcpy ((void *)&sttynew, (void *)&sttyold, sizeof(sttynew));

    /* Setup new settings for Raw mode */
    sttynew.c_iflag = IGNBRK;
    sttynew.c_oflag = 0;
    cflag = sttyold.c_cflag;
    /* Clear items in cflag that we will be explicitly setting.     */
    cflag &= ~ (CBAUD | PARENB | PARODD | CSIZE);
    switch (baud) {                     /* Set baudrate.            */
    case   300: cflag |= B300;   break;
    case  2400: cflag |= B2400;  break;
    case  4800: cflag |= B4800;  break;
    case  9600: cflag |= B9600;  break;
    case 19200: cflag |= B19200; break;
    case 38400: cflag |= B38400; break;
    default:
        TIMESTAMP(stderr,qnow_gm());
        fprintf(stderr, "Invalid baudrate %d\n", baud);
        close (fd);
        return (-1);
    }
    switch (parity) {                   /* Set parity.              */
    case 'e':
    case 'E':           cflag |= PARENB; break;
    case 'o':
    case 'O':           cflag |= PARENB | PARODD; break;
    case 'n':
    case 'N':           break;
    default:
        TIMESTAMP(stderr,qnow_gm());
        fprintf(stderr, "Invalid parity %c\n", parity);
        close (fd);
        return (-1);
    }
    cflag |= CS8;                       /* Set character size.      */
    cflag |= CREAD;                     /* Enable receiver.         */
    sttynew.c_cflag = cflag;
    sttynew.c_lflag = NOFLSH;
    sttynew.c_cc[VMIN] = 0;             /* Min characters.          */
#ifdef  POLL
    sttynew.c_cc[VTIME] = 0;            /* Timeout value.           */
#else	/* POLL */
    sttynew.c_cc[VTIME] = 1 * unix_service_time;    /* Timeout value.*/
#endif	/* POLL */

    /* Configure port using new settings */
#ifdef  TERMIO
    if (ioctl(fd, TCSETA, &sttynew) < 0) {
        TIMESTAMP(stderr,qnow_gm());
        fprintf (stderr, "Error configuring line %s\n", line);
        close (fd);
        return (-1);
    }
#else	/* TERMIO */
    if (tcsetattr(fd, TCSANOW, &sttynew) < 0) {
        TIMESTAMP(stderr,qnow_gm());
        fprintf (stderr, "Error configuring line %s\n", line);
        close (fd);
        return (-1);
    }
#endif	/* TERMIO */
    if (DEBUG(DEBUG_INPUT)) {
        fprintf (info, "Set line attributes for %s\n", line);
    }

    /* Raise RTS so that Ashtech receiver will output data.             */
/*::
    if (ioctl (fd, TIOCMGET, &mflags) < 0) {
	TIMESTAMP(stderr,qnow_gm());
	fprintf (stderr, "Error getting modem line settings\n");
	close (fd);
	return (-1);
    }
::*/
    bis = TIOCM_RTS | TIOCM_DTR;
    if (ioctl (fd, TIOCMBIS, &bis) < 0) {
        TIMESTAMP(stderr,qnow_gm());
        fprintf (stderr, "Error setting modem line settings\n");
        close (fd);
        return (-1);
    }

    serial_io = 1;
    return (fd);
}

#else	/* _OSK */
/************************************************************************/
/*  open_serial -                                                       */
/*      Open a serial line for reading.                                 */
/************************************************************************/
int open_serial (char *line, int baud, char parity)
{
    int fd;
    int mflags, bis;
    int status = 0;
    struct sgbuf sttynew ;
    struct sgbuf sttyold ;

    unsigned int cflag = 0;

    if ((fd = open (line, (short int)(S_IREAD|S_IWRITE))) < 0) {
        TIMESTAMP(stderr,qnow_gm());
        fprintf (stderr, "Error opening line %s\n", line);
        return (-1);
    }
    if (DEBUG(DEBUG_INPUT)) {
        fprintf (info, "Opened serial line %s fd=%d\n", line, fd);
    }
    if (_gs_opt(fd, &sttyold) == ERROR) {
        fprintf (stderr, "Could not obtain settings for %s\n", line);
        exit (12);
    }
    memcpy ((char *) &sttynew, (char *) &sttyold, sizeof(struct sgbuf));

    /* Setup new settings for Raw mode */
    memset((char *) &sttynew.sg_case, '\0', 19) ; /* clear line editing */
    memset((char *) &sttynew.sg_xon, '\0', 4) ; /* clear flow and tabs */
    switch (baud) {
    case 2400:  sttynew.sg_baud = 0xA; break;
    case 4800:  sttynew.sg_baud = 0xC; break;
    case 9600:  sttynew.sg_baud = 0xE; break;
    case 19200: sttynew.sg_baud = 0xF; break;
    case 38400: sttynew.sg_baud = 6; break;
    default:            
        fprintf(stderr, "Invalid baudrate %d\n", baud);
        exit(12) ;
    }
    cflag = 0 ; /* assume no parity */
    switch (parity) {
    case 'e':
    case 'E':           cflag = 3; break;
    case 'o':
    case 'O':           cflag = 1; break;
    case 'n':
    case 'N':           break;
    default:
        TIMESTAMP(stderr,qnow_gm());
        fprintf(stderr, "Invalid parity %c\n", parity);
        close (fd);
        return (-1);
    }
    sttynew.sg_parity = (sttynew.sg_parity & 0xfc) | cflag ;

    /* Configure port using new settings */
    if (_ss_opt(fd, &sttynew) == ERROR) {
        fprintf (stderr, "Could not configure port %s\n", line) ;
        exit (12) ;
    }
    if ((status = _ss_enrts (fd)) == -1) {
        fprintf (stderr, "Could not enable RTS on fd %d port %s, errno=%d ", 
                 fd, line, errno) ;
        fprintf (stderr, "status=%d\n", status);
    }
    if ((status = resetp (fd, 4096)) != 0) {
        fprintf (stderr, "Could not change serial input buffer size on fd %d port %s ", 
                 fd, line) ;
        fprintf (stderr, "status=%d\n", status);
    }
    if (DEBUG(DEBUG_INPUT)) {
        fprintf (info, "Set line attributes for %s\n", line);
    }

    serial_io = 1;
    return (fd);
}
#endif	/* _OSK */
