/************************************************************************/
/*  File:       trimble_bin.h                                           */
/*                                                                      */
/*  Description:                                                        */
/*      definitions and structures for binary Trimble data.             */
/*                                                                      */
/*      Ray Baxter                                                      */
/*      Seismological Laboratory                                        */
/*      University of California, Berkeley                              */
/*      ray@seismo.berkeley.edu                                         */
/*                                                                      */
/************************************************************************/

/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/*
Modification History:
------------------------------------------------------------------------
95/11/13 DSN    1.0     Initial coding.
96/06/16 DSN    1.3     Add RTR error record.
96/10/01 DSN    1.4     Added copyright msg.
97/11/26 DRB	1.5     Adapted to Trimble.
*/

/************************************************************************/

/*      @(#)trimble_bin.h       1.1 09/09/97 19:04:10   */

#define UNKNOWN_REC      -1             /* Unknown record id.           */
#define SVDATA         0x55             /* Ephemeris, Almanac & Flags   */
#define RAWDATA        0x57             /* Real-time Survey Data        */

#define MAX_SVS			 12
#define SV_FLAGS          1             /* Data type indicator values.  */
#define SV_EPHEMERIS      2             /*   This is from byte five of  */
#define SV_ALMANAC        3             /*   an SVDATA record.          */ 
#define SV_ION_UTC        4             /*                              */

/**************   THESE PROBABLY HAVE TO BE INCREASED 16x  **************/

#define MAXRLEN         256             /* Max record len in chars.     */
#define MAXSLEN         128             /* Max record len in short ints.*/

#define NUL 0
#define STX 2
#define ETX 3

#define SVDATA_PER_SEC    1             /* 1/svdata units (secs)        */
#define RAWDATA_PER_SEC 1000            /* 1/rawdata units (secs)       */

#define CONCISE_RECORD  0x01            /* Concise rawdata record.      */
#define EXPANDED_RECORD 0x02            /* Expanded rawdata record.     */

typedef struct _SV_FLAGS_STRUCT {       /* SV flags from svdata record. */
  long eph_flags;           
  long alm_flags;
  long sv_disabled_flags;
  long sv_unhealthy_flags;
  long l1_flags;
  long l2_flags;
  long ycode_flags;
  long pcode_flags;
  long r1;                              /* Reserved by manufacturer.     */
  long r2;                              /* Reserved by manufacturer.     */
  long r3;                              /* Reserved by manufacturer.     */
  long r4;                              /* Reserved by manufacturer.     */
} SV_FLAGS_STRUCT;

typedef struct _eph_data {              /* Ephemeris data from svdata.   */
  unsigned short week;
  unsigned short iodc;
  unsigned char r1;                     /* Reserved by manufacturer.     */
  unsigned char iode;
  long tow;
  long toc;
  long toe;
  double tgd;
  double af2;
  double af1;
  double af0;
  double crs;
  double delta_n;
  double m_sub_0;
  double cuc;
  double eccentricity;
  double cus;
  double sqrt_a;
  double cic;
  double omega_sub_0;
  double cis;
  double i_sub_0;
  double crc;
  double omega;
  double omega_dot;
  double i_dot;
  long flags;
} EPH_DATA;

typedef struct _almanac {        /* Almanac data from svdata record.    */
  unsigned long alm_decode_time;
  short awn;
  unsigned long toa;
  double sqrta;
  double eccent;
  double isubo;
  double omega_dot;
  double omega_sub_0;
  double omega;
  double m_sub_0;
  unsigned char alm_health;
} ALMANAC;

typedef struct _ion_utc {       /* ION/UTC data from svdata record.     */
  double alpha_0;               /* UTC Parameters.                      */
  double alpha_1;
  double alpha_2;
  double alpha_3;
  double beta_0;
  double beta_1;
  double beta_2;
  double beta_3;
  double a_sub_0;               /* ION Parameters.                      */
  double a_sub_1;
  double t_sub_0T;
  double delta_t_ls;
  double delta_t_lsf;
  double iontime;
  unsigned char wn_sub_t;
  unsigned char wn_sub_lsf;
  unsigned char dn;
  unsigned char r[6];           /* Reserved by Trimble.                 */
} ION_UTC;

typedef struct _svdata {        /* Whole SVDATA record from receiver.   */
  unsigned char stx;  
  unsigned char status;
  unsigned char type;
  unsigned char length;
  unsigned char datatype;
  unsigned char sv_prn_num;
  union {
    SV_FLAGS_STRUCT sv_flags;
    EPH_DATA eph_data;
    ALMANAC  almanac;
    ION_UTC  ion_utc;
  } data;
  unsigned char checksum;
  unsigned char etx;
} SVDATA_STRUCT;

typedef struct _con_sv_block {  /* SV info in concise format.           */
  unsigned char prn;
  unsigned char flags1;
  unsigned char flags2;
  char ele_angle;
  int azimuth;
  unsigned char l1_snr;         /* Beginning of L1 Data.                */
  double l1_p_range;            /*   Non 0 if bit 6 of flags1 is set.   */
  double l1_continuous_phase;
  float l1_doppler;
  unsigned char l2_snr;         /* Beginning of L2 data.                */
  double l2_continuous_phase;   /*   Availible if bit 0 of flags1 set.  */
  float l2_p_range;
  unsigned char iode;           /* Real-time data. Flags.               */
  unsigned char l1_slips;       /*   Avaiible if RECORD INTERPRETATION. */
  unsigned char l2_slips;
} CON_SV_BLOCK;

typedef struct _exp_sv_block {        /* */
	unsigned char sv_prn_num;
	unsigned char flags1;
	unsigned char flags2;
	unsigned char flag_status;
	int elevation_angle;
	int azimuth;
  	unsigned char l1_snr;         /* Beginning of L1 Data.                */
  	double l1_p_range;            /*   Non 0 if bit 6 of flags1 is set.   */
  	double l1_continuous_phase;
  	double l1_doppler;
  	double l1_reserved;
  	double l2_snr;         	      /* Beginning of L2 data.                */
  	double l2_continuous_phase;   /*   Availible if bit 0 of flags1 set.  */
  	double l2_p_range;
  	unsigned char iode;           /* Real-time data. Flags.               */
  	unsigned char l1_slips;       /*   Avaiible if RECORD INTERPRETATION. */
  	unsigned char l2_slips;
  	unsigned char r1;	      /* Reserved by Trimble. */
  	double l2_doppler;
} EXP_SV_BLOCK;

typedef struct _position {
	double latitude;
	double longitude;
	double altitude;
	double clock_offset;
	double frequency_offset;
	double pdop;
	double latitude_rate;
	double longitude_rate;
	double altitude_rate;
	unsigned long gps_msec_of_week;
	unsigned char position_flags;
	unsigned char n_svs;
	unsigned char channel[MAX_SVS];
	unsigned char prn_number[MAX_SVS];
} POSITION;

typedef struct _concise {       /* Concise formatted data from RAWDATA. */
  double receive_time;
  double clock_offset;
  unsigned char n_svs;
  CON_SV_BLOCK con_sv_block[MAX_SVS];
} CONCISE;

typedef struct _expanded {      /* Expanded format data from RAWDATA.   */
  double receive_time;
  double clock_offset;
  unsigned char n_svs;
  EXP_SV_BLOCK exp_sv_block[MAX_SVS];
  POSITION position;
} EXPANDED;

typedef union _con_or_exp {
  CONCISE concise;
  EXPANDED expanded;
} CON_OR_EXP;

typedef struct _RAWDATA {
  unsigned char stx;
  unsigned char status;
  unsigned char type;
  unsigned char length;
  unsigned char record_type;
  unsigned char page_counter;
  unsigned char reply_number;
  unsigned char record_interpretation;
  CON_OR_EXP con_or_exp;
  unsigned char checksum;
  unsigned char etx;
} RAWDATA_STRUCT;

typedef union _trimble_record {
  RAWDATA_STRUCT rawdata;
  SVDATA_STRUCT svdata;
} TRIMBLE_RECORD;

