/************************************************************************/
/*  Interface routines for Trimble 4000 SE/SSE/SSi GPS receivers.       */
/*                                                                      */
/*      Ray Baxter                                                      */
/*      Seismological Laboratory                                        */
/*      University of California, Berkeley                              */
/*      ray@seismo.berkeley.edu                                         */
/*                                                                      */
/************************************************************************/

/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for educational, research and non-profit purposes,
 * without fee, and without a written agreement is hereby granted,
 * provided that the above copyright notice, this paragraph and the
 * following three paragraphs appear in all copies.
 * 
 * Permission to incorporate this software into commercial products may
 * be obtained from the Office of Technology Licensing, 2150 Shattuck
 * Avenue, Suite 510, Berkeley, CA  94704.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 * INCLUDING LOST PROFITS, ARISING OUT OF THE USE OF THIS SOFTWARE AND
 * ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF CALIFORNIA HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
 * CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

/************************************************************************/

#ifndef lint
static char sccsid[] = "@(#)trimble_subs.c      1.1 09/09/97 19:03:42";
#endif	/* lint */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

#ifndef _OSK
#include <sys/types.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <qlib2.h>
#else	/* _OSK */
#include <types.h>
#include <sysio.h>
#define strrchr rindex
/* time.h included by stdlib.h */
#endif	/* _OSK */

#include "gps_generic.h"
#include "receiver_subs.h"
#include "serial_io.h"
#include "trimble_bin.h"

/************************************************************************/
/*  Vendor-specific definitions.                                        */
/************************************************************************/
#define DEFAULT_PATTERN "%s%j0.d%H"
char gps_vendor[120]   = "TRIMBLE";
char gps_model[120]    = "";
char gps_software[120] = "";
char gps_firmware[120] = "";

/************************************************************************/
/*  Externals                                                           */
/************************************************************************/
/*                                                                      */
/*  Allocate buffer for use in main procecure, since the size of the    */
/*  records is receiver dependent.                                      */
/************************************************************************/
char input_rec[MAXRLEN];        /* buffer to hold input binary records. */
int maxrlen = MAXRLEN;          /* max input record length.             */
char out_pattern[] = DEFAULT_PATTERN;   /* outfile pattern template.    */
extern FILE *info;              /* output file ptr for information.     */
extern int debug_option;        /* default debug option.                */
extern int verbosity;           /* verbosity option.                    */
#ifndef _OSK
extern FILE *out;               /* output file ptr.                     */
#else	/* _OSK */
extern int out;                 /* output file descriptor.              */
#endif	/* _OSK */

/************************************************************************/
/*  Static variables.                                                   */
/************************************************************************/
static int decimate;
static int sample_interval;

static int rawdata_interval;
static int svdata_interval;

static TRIMBLE_RECORD u;

/************************************************************************/
/*  Internal function declarations.                                     */
/************************************************************************/
char *record_string (int record_type);
int end_of_epoch (int record_type, void *record);
int get_bin_record (char *line, int maxlen, int *rec_type, int *rec_len, int *nskip, int fd);
int getbyte( int fd );
int parse_svdata ( unsigned char *line, SVDATA_STRUCT *svdata);
int process_record (int rec_type, int rec_len, char *line);
int setup_decimation (int decimate_flag, int sample_interval_val);
void dump_almanac ( ALMANAC *almanac, FILE *fp);
void dump_concise_data ( CONCISE *concise, FILE *info );
void dump_expanded_data ( EXPANDED *expanded, FILE *info );
void dump_ion_utc ( ION_UTC *ion_utc, FILE *fp);
void dump_rawdata ( RAWDATA_STRUCT *rawdata, FILE *info );
void dump_sv_ephemeris ( EPH_DATA *eph_data, FILE *fp);
void dump_svdata (SVDATA_STRUCT *svdata, FILE *fp);
void dump_svflags ( SV_FLAGS_STRUCT *sv_flags, FILE *fp);
void parse_con_sv_block ( unsigned char *line, CON_SV_BLOCK *con_block );
void parse_concise_data ( unsigned char *line,  CONCISE *concise );
void parse_exp_sv_block ( unsigned char * line,  EXP_SV_BLOCK *exp_block );
void parse_expanded_data ( unsigned char *line,  EXPANDED *expanded );
void parse_position ( unsigned char *line,  POSITION *position );
void parse_rawdata ( unsigned char *line, RAWDATA_STRUCT *rawdata );
void parse_sv_almanac ( unsigned char *line, ALMANAC *almanac);
void parse_sv_ephemeris ( unsigned char *line, EPH_DATA *eph_data);
void parse_sv_flags ( unsigned char *line, SV_FLAGS_STRUCT *flags);
void parse_sv_ion_utc ( unsigned char *line, ION_UTC *ion_utc);
void putbytes( int nbytes );

/************************************************************************/
/*  setup_decimation:                                                   */
/*      Setup for decimation.                                           */
/*  Return:                                                             */
/*      0 on OK, -1 on error.                                           */
/************************************************************************/
int setup_decimation (int decimate_flag, int sample_interval_val)
{
    decimate = decimate_flag;
    sample_interval = sample_interval_val;
    if (decimate && sample_interval <= 0.0) {
        fprintf (stderr, "Error - sample interval must be > 0\n");
        return (-1);
    }
    if (decimate) {
	rawdata_interval = sample_interval * RAWDATA_PER_SEC;
	svdata_interval = sample_interval * SVDATA_PER_SEC;
    }
    return (0);
}
    
/************************************************************************/
/*  get_bin_record -                                                    */
/*      read, verify, and determine type of binary record.              */
/************************************************************************/
int get_bin_record (char *line, int maxlen, int *rec_type, int *rec_len, int *nskip, int fd)
{
    int c;
    int i;
    int data_len;
    int length;
    int cksum;
    int found_cksum;

    /* Frame on header  */

    if (verbosity >= 1) fprintf (info, "Enter get_bin_record\n");
    *nskip = 0;

    while (1) {
	*rec_type = UNKNOWN_REC;
	while ( (c = getbyte(fd) ) != STX ) {
	    if ( c == EOF ) {
		if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n", EOF);
		return(EOF);
	    }
	    ++*nskip;
	}
	/* We have an STX. */
	line[0] = c;
	length = 1;

	if( (c = getbyte(fd)) != NUL ) { 
	    if ( c == EOF ) {
		if ( verbosity >= 1 ) fprintf (info, "Exit get_bin_record rc=%d\n", EOF);
		return(EOF);
	    }
	    putbytes(1);
	    continue;
	}
	/* We have a NUL (OK) Status Byte. */

	line[1] = c;
	length = 2;

	c = getbyte(fd);
	if( c != SVDATA && c != RAWDATA ) {
	    if ( c == EOF ) {
		if ( verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n", EOF);
		return(EOF);
	    }
	    putbytes(2);
	    continue;
	}
	/* We have one of the record types we know, SVDATA or RAWDATA. */
	line[2] = *rec_type = c; 
	length = 3;

	c = getbyte(fd);
	if ( c == EOF ) {
	    if ( verbosity > 1 ) fprintf (info, "Exit get_bin_record rc=%d\n", EOF );
	    return(EOF);
	}
	/* This is the length of the data section of this record. */
	data_len = c;
	line[3] = c;
	length = 4;

	*rec_len = c + 6;

	if (DEBUG(DEBUG_INPUT)) {
	    fprintf (info, "Header found\n");
	}
        
	/* Read the data section of the record. */
    
	/* Initalize the checksum, defined in DCFSM 1-2. */
	cksum = line[1] + line[2] + line[3];
    
	for (i=4; i < data_len+4; i++) {
	    c = getbyte(fd);
	    if (c==EOF) {
		if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n", EOF);
		*rec_len = i;
		return(EOF);
	    }
	    line[i] = c;
	    ++length;
	    cksum += c;
	}
    
	/* Get the checksum value. */
	c = getbyte(fd); 
	if (c==EOF) {
	    if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n", EOF);
	    *rec_len = i;
	    return(EOF);
	}
	line[i] = found_cksum = c;
	++length;
	++i;

	/* Get the ending ETX. */
	c = getbyte(fd);
	if (c==EOF) {
	    if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n", EOF);
	    return(EOF);
	}
	if ( c != ETX ) {
	    if (DEBUG(DEBUG_INPUT)) {
		fprintf ( info, "No ETX at end of record.\n" );
	    }

	    putbytes(*rec_len);
	    continue;
	}
	line[i] = c;
	++length;

	if (length != *rec_len ) {
	    fprintf ( stderr, "Values for record length differ.\n" );
	}
	/* This is a complete record, so set the variables. */
	cksum %= 256;

	if (cksum != found_cksum) {
	    if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=%d\n", CKSUM_ERROR);
	    return (CKSUM_ERROR);
	}

	/* Return record type;                                      */

	if (verbosity >= 1) fprintf (info, "Exit get_bin_record rc=0\n");
	return (RECORD_OK);
    }
}

/************************************************************************/
/*  process_record:                                                     */
/*      Process input record.                                           */
/*  Return:                                                             */
/*      Negative value on error.                                        */
/*      Non-negative value on record OK.                                */
/*      0:  output record.                                              */
/*      1:  skip record (decimation or uninteresting record).           */
/************************************************************************/
int process_record (int rec_type, int rec_len, char *line)
{
    int status = RECORD_OK;
    int is_svdata_rec = 0;
    int is_rawdata_rec = 0;
    if (verbosity >= 1) fprintf (info, "Enter process_record\n");

    switch (rec_type) {
    case SVDATA:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Record SVDATA\n");
        parse_svdata ( (unsigned char *)line, &u.svdata );
        if (DEBUG(DEBUG_RECORD)) dump_svdata ( &u.svdata, info );
        is_svdata_rec = 1;
        break;

    case RAWDATA:
        if (DEBUG(DEBUG_TYPE)) fprintf (info, "Record RAWDATA\n");
        parse_rawdata ( (unsigned char *)line, &u.rawdata);
        if (DEBUG(DEBUG_RECORD)) dump_rawdata ( &u.rawdata, info);
        is_rawdata_rec = 1;
        break;
    default:
        fprintf (info, "Unknown record type %d\n", rec_type);
        status = RECORD_ERROR;
        break;
    }

    /* Perform decimation if requested.                         */

/* Decimation not implemented. 
 *

 if (decimate && is_svdata_rec && (int)u.svdata.seq_tag % (svdata_interval) != 0) {
 status = SKIP_RECORD;
 }
 if (decimate && is_rawdata_rec && (int)u.rawdata.pbentime % (rawdata_interval) != 0) {
 status = SKIP_RECORD;
 }
 *
 */

    if (verbosity >= 1) fprintf (info, "Exit process_record rc=0\n");

    return (status);
}

/************************************************************************/
/*  parse_svdata -                                                        */
/*      Parse an svdata record.                                              */
/************************************************************************/
int parse_svdata ( unsigned char *line, SVDATA_STRUCT *svdata)
{
    svdata->stx = line[0];
    svdata->status = line[1];
    svdata->type = line[2];
    svdata->length = line[3];
    svdata->datatype = line[4];
    svdata->sv_prn_num = line[5];
    svdata->checksum = line[svdata->length+4];
    svdata->etx = line[svdata->length+5];

    switch (svdata->datatype) {
    case SV_FLAGS:
        parse_sv_flags ( line+6, &svdata->data.sv_flags);
        break;
    case SV_EPHEMERIS:
        parse_sv_ephemeris ( line+6, &svdata->data.eph_data);
        break;
    case SV_ALMANAC:
        parse_sv_almanac ( line+6, &svdata->data.almanac);
        break;
    case SV_ION_UTC:
        parse_sv_ion_utc ( line+6, &svdata->data.ion_utc);
        break;
    default:
#ifndef _OSK
        fprintf (stderr, "Unknown SVDATA type %d\n", svdata->datatype);
#else	/* _OSK */
        fprintf (stderr, "Unknown SVDATA type %d\n", svdata->datatype);
#endif	/* _OSK */
        return (FORMAT_ERROR);
    }
    return (0);
}

void parse_sv_flags ( unsigned char *line, SV_FLAGS_STRUCT *flags)
{
    memcpy ((void *)&flags->eph_flags, (void *)&line[0], 4);           
    memcpy ((void *)&flags->alm_flags, (void *)&line[4], 4);
    memcpy ((void *)&flags->sv_disabled_flags, (void *)&line[8], 4);
  
    memcpy ((void *)&flags->sv_unhealthy_flags, (void *)&line[12], 4);
    memcpy ((void *)&flags->l1_flags, (void *)&line[16], 4);
    memcpy ((void *)&flags->l2_flags, (void *)&line[20], 4);
  
    memcpy ((void *)&flags->ycode_flags, (void *)&line[24], 4);
    memcpy ((void *)&flags->pcode_flags, (void *)&line[28], 4);
    memcpy ((void *)&flags->r1, (void *)&line[32], 4);
  
    memcpy ((void *)&flags->r2, (void *)&line[36], 4);
    memcpy ((void *)&flags->r3, (void *)&line[40], 4);
    memcpy ((void *)&flags->r4, (void *)&line[44], 4);
  
    return;
}

void parse_sv_ephemeris ( unsigned char *line, EPH_DATA *eph_data)
{
    memcpy ((void *)&eph_data->week, (void *)&line[0], 2);
    memcpy ((void *)&eph_data->iodc, (void *)&line[2], 2);
    memcpy ((void *)&eph_data->r1, (void *)&line[5], 1);
  
    memcpy ((void *)&eph_data->iode, (void *)&line[5], 1);
    memcpy ((void *)&eph_data->tow, (void *)&line[6], 4);
    memcpy ((void *)&eph_data->toc, (void *)&line[10], 4);
  
    memcpy ((void *)&eph_data->toe, (void *)&line[14], 4);
    memcpy ((void *)&eph_data->tgd, (void *)&line[18], 8);
    memcpy ((void *)&eph_data->af2, (void *)&line[26], 8);
  
    memcpy ((void *)&eph_data->af1, (void *)&line[34], 8);
    memcpy ((void *)&eph_data->af0, (void *)&line[42], 8);
    memcpy ((void *)&eph_data->crs, (void *)&line[50], 8);
  
    memcpy ((void *)&eph_data->delta_n, (void *)&line[58], 8);
    memcpy ((void *)&eph_data->m_sub_0, (void *)&line[64], 8);
    memcpy ((void *)&eph_data->cuc, (void *)&line[72], 8);
  
    memcpy ((void *)&eph_data->eccentricity, (void *)&line[80], 8);
    memcpy ((void *)&eph_data->cus, (void *)&line[88], 8);
    memcpy ((void *)&eph_data->sqrt_a, (void *)&line[96], 8);
  
    memcpy ((void *)&eph_data->cic, (void *)&line[104], 8);
    memcpy ((void *)&eph_data->omega_sub_0, (void *)&line[112], 8);
    memcpy ((void *)&eph_data->cis, (void *)&line[120], 8);
  
    memcpy ((void *)&eph_data->i_sub_0, (void *)&line[128], 8);
    memcpy ((void *)&eph_data->crc, (void *)&line[136], 8);
    memcpy ((void *)&eph_data->omega, (void *)&line[144], 8);
  
    memcpy ((void *)&eph_data->omega_dot, (void *)&line[152], 8);
    memcpy ((void *)&eph_data->i_dot, (void *)&line[160], 8);
    memcpy ((void *)&eph_data->flags, (void *)&line[168], 4);

    return;
}

void parse_sv_almanac ( unsigned char *line, ALMANAC *almanac)
{
    memcpy ( (void *)&almanac->alm_decode_time, (void *)&line[0], 4);
    memcpy ( (void *)&almanac->awn, (void *)&line[4], 2);
    memcpy ( (void *)&almanac->toa, (void *)&line[6], 4);
  
    memcpy ( (void *)&almanac->sqrta, (void *)&line[10], 8);
    memcpy ( (void *)&almanac->eccent, (void *)&line[18], 8);
    memcpy ( (void *)&almanac->isubo, (void *)&line[26], 8);
  
    memcpy ( (void *)&almanac->omega_dot, (void *)&line[34], 8);
    memcpy ( (void *)&almanac->omega_sub_0, (void *)&line[42], 8);
    memcpy ( (void *)&almanac->omega, (void *)&line[50], 8);
  
    memcpy ( (void *)&almanac->m_sub_0, (void *)&line[58], 8);
    memcpy ( (void *)&almanac->alm_health, (void *)&line[66], 1);

    return;
}

void parse_sv_ion_utc ( unsigned char *line, ION_UTC *ion_utc)
{
    memcpy ( (void *)&ion_utc->alpha_0, (void *)&line[0], 8);            
    memcpy ( (void *)&ion_utc->alpha_1, (void *)&line[8], 8);
    memcpy ( (void *)&ion_utc->alpha_2, (void *)&line[16], 8);
  
    memcpy ( (void *)&ion_utc->alpha_3, (void *)&line[24], 8);
    memcpy ( (void *)&ion_utc->beta_0, (void *)&line[32], 8);
    memcpy ( (void *)&ion_utc->beta_1, (void *)&line[40], 8);
  
    memcpy ( (void *)&ion_utc->beta_2, (void *)&line[48], 8);
    memcpy ( (void *)&ion_utc->beta_3, (void *)&line[56], 8);
    memcpy ( (void *)&ion_utc->a_sub_0, (void *)&line[64], 8);
  
    memcpy ( (void *)&ion_utc->a_sub_1, (void *)&line[72], 8);
    memcpy ( (void *)&ion_utc->t_sub_0T, (void *)&line[80], 8);
    memcpy ( (void *)&ion_utc->delta_t_ls, (void *)&line[88], 8);
  
    memcpy ( (void *)&ion_utc->delta_t_lsf, (void *)&line[96], 8);
    memcpy ( (void *)&ion_utc->iontime, (void *)&line[104], 8);
    memcpy ( (void *)&ion_utc->wn_sub_t, (void *)&line[112], 8);
  
    memcpy ( (void *)&ion_utc->wn_sub_lsf, (void *)&line[120], 1);
    memcpy ( (void *)&ion_utc->dn, (void *)&line[121], 1);
    memcpy ( (void *)&ion_utc->r[6], (void *)&line[122], 6);

    return;
}

/************************************************************************/
/*  dump_svdata -                                                       */
/*      Dump SVDATA record.                                             */
/************************************************************************/
void dump_svdata (SVDATA_STRUCT *svdata, FILE *fp)
{
    fprintf (fp, "SVDATA %-15.15s = %02x\n", "stx", svdata->stx);
    fprintf (fp, "SVDATA %-15.15s = %02x\n", "status", svdata->status);
    fprintf (fp, "SVDATA %-15.15s = %02x\n", "type", svdata->type);
    fprintf (fp, "SVDATA %-15.15s = %02x\n", "length", svdata->length);
    fprintf (fp, "SVDATA %-15.15s = %02x\n", "datatype", svdata->datatype);
    fprintf (fp, "SVDATA %-15.15s = %02x\n", "sv_prn_num", svdata->sv_prn_num);
    switch (svdata->datatype) {
    case SV_FLAGS:
	dump_svflags (&svdata->data.sv_flags, fp);
	break;
    case SV_EPHEMERIS:
	dump_sv_ephemeris (&svdata->data.eph_data, fp);
	break;
    case SV_ALMANAC:
	dump_almanac (&svdata->data.almanac, fp);
	break;
    case SV_ION_UTC:
	dump_ion_utc (&svdata->data.ion_utc, fp);
	break;
    default:
#ifndef _OSK
	fprintf (fp, "Unknown SVDATA type %02x\n", svdata->type);
#else	/* _OSK */
	fprintf (fp, "Unknown SVDATA type %02x\n", svdata->type);
#endif	/* _OSK */
	return;
    }

    fprintf (fp, "SVDATA %-15.15s = %02x\n", "checksum", svdata->checksum);   
    fprintf (fp, "SVDATA %-15.15s = %02x\n", "etx", svdata->etx);

    fprintf (fp, "\n");
    return;
}

void dump_svflags ( SV_FLAGS_STRUCT *sv_flags, FILE *fp)
{
/* I copied this from Doug, but check on the strings passed to fprintf */
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "eph_flags", sv_flags->eph_flags);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "alm_flags", sv_flags->alm_flags);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "sv_disabled_flags", sv_flags->sv_disabled_flags);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "sv_unhealthy_flags", sv_flags->sv_unhealthy_flags);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "l1_flags", sv_flags->l1_flags);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "l2_flags", sv_flags->l2_flags);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "ycode_flags", sv_flags->ycode_flags);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "pcode_flags", sv_flags->pcode_flags);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "r1", sv_flags->r1);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "r2", sv_flags->r2);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "r3", sv_flags->r3);
    fprintf (fp, "SV FLAGS %-15.15s = %ld\n", "r4", sv_flags->r4);
    fprintf (fp, "\n");
    return;
}

void dump_sv_ephemeris ( EPH_DATA *eph_data, FILE *fp)
{
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %d\n", "week", eph_data->week);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %d\n", "iodc", eph_data->iodc);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %d\n", "r1", eph_data->r1);
  
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %d\n", "iode", eph_data->iode);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %ld\n", "tow", eph_data->tow);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %ld\n", "toc", eph_data->toc);
  
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %ld\n", "toe", eph_data->toe);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->tgd);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->af2);
  
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->af1);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->af0);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->crs);
  
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->delta_n);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->m_sub_0);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->cuc);
  
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->eccentricity);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->cus);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->sqrt_a);
  
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->cic);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->omega_sub_0);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->cis);
  
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->i_sub_0);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->crc);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->omega);
  
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->omega_dot);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %g\n", "toe", eph_data->i_dot);
    fprintf (fp, "SV_EPHEMERIS %-15.15s = %ld\n", "toe", eph_data->flags); 
  
    return;
}

void dump_almanac ( ALMANAC *almanac, FILE *fp)
{
    fprintf (fp, "SV_ALMANAC %-15.15s = %ld\n", "alm_decode_time", almanac->alm_decode_time );
    fprintf (fp, "SV_ALMANAC %-15.15s = %d\n", "awn", almanac->awn );
    fprintf (fp, "SV_ALMANAC %-15.15s = %ld\n", "toa", almanac->toa );
  
    fprintf (fp, "SV_ALMANAC %-15.15s = %g\n", "sqrta", almanac->sqrta );
    fprintf (fp, "SV_ALMANAC %-15.15s = %g\n", "eccent", almanac->eccent );
    fprintf (fp, "SV_ALMANAC %-15.15s = %g\n", "isubo", almanac->isubo );
  
    fprintf (fp, "SV_ALMANAC %-15.15s = %g\n", "omega_dot", almanac->omega_dot );
    fprintf (fp, "SV_ALMANAC %-15.15s = %g\n", "omega_sub_0", almanac->omega_sub_0 );
    fprintf (fp, "SV_ALMANAC %-15.15s = %g\n", "omega", almanac->omega );
  
    fprintf (fp, "SV_ALMANAC %-15.15s = %g\n", "m_sub_0", almanac->m_sub_0 );
    fprintf (fp, "SV_ALMANAC %-15.15s = %c\n", "alm_health", almanac->alm_health );

    return;
}
void dump_ion_utc ( ION_UTC *ion_utc, FILE *fp)
{
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "alpha_0", ion_utc->alpha_0 );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "alpha_1", ion_utc->alpha_1 );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "alpha_2", ion_utc->alpha_2 );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "alpha_3", ion_utc->alpha_3 );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "beta_0", ion_utc->beta_0 );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "beta_1", ion_utc->beta_1 );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "beta_2", ion_utc->beta_2 );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "beta_3", ion_utc->beta_3 );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "a_sub_1", ion_utc->a_sub_1 );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "t_sub_0T", ion_utc->t_sub_0T );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "delta_t_ls", ion_utc->delta_t_ls );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "delta_t_lsf", ion_utc->delta_t_lsf );
    fprintf (fp, "SV_ION_UTC %-15.15s = %g\n", "iontime", ion_utc->iontime );
    fprintf (fp, "SV_ION_UTC %-15.15s = %c\n", "wn_sub_t", ion_utc->wn_sub_t );
    fprintf (fp, "SV_ION_UTC %-15.15s = %c\n", "wn_sub_lsf", ion_utc->wn_sub_lsf );
    fprintf (fp, "SV_ION_UTC %-15.15s = %c\n", "dn", ion_utc->dn );
    fprintf (fp, "SV_ION_UTC %-15.15s = %6s\n", "r", ion_utc->r );
    return;
}

void parse_rawdata ( unsigned  char *line, RAWDATA_STRUCT *rawdata )
{
    int concise, expanded;
  
    rawdata->stx = line[0];
    rawdata->status = line[1];
    rawdata->type = line[2];
    rawdata->length = line[3];
    rawdata->record_type = line[4];
    rawdata->page_counter = line[5];
    rawdata->reply_number = line[6];
    rawdata->record_interpretation = line[7];
  
    rawdata->checksum = line[rawdata->length+4];
    rawdata->etx = line[rawdata->length+5];

    concise = rawdata->record_interpretation & CONCISE_RECORD; 
    expanded = rawdata->record_interpretation & EXPANDED_RECORD;
        
    if (concise && expanded ) {
	fprintf ( stderr, "A record was detected that contained both concise and expanded raw data. \n" );
	return;
    } else if (concise ) {
	parse_concise_data ( &line[8], &rawdata->con_or_exp.concise );
    }
    else if ( expanded ) {
	parse_expanded_data ( &line[8], &rawdata->con_or_exp.expanded );
	fprintf ( stderr, "Unknown record type, neither concise nor expanded rawdata.\n" );
	return;
    }

    return;
}

void parse_concise_data ( unsigned char *line,  CONCISE *concise )
{
    int i;

    memcpy ( (void *)&concise->receive_time, (void *)&line[0], 8);
    memcpy ( (void *)&concise->clock_offset, (void *)&line[8], 8);
    memcpy ( (void *)&concise->n_svs, (void *)&line[16], 1);

    for ( i = 0; i < concise->n_svs ; i++ )
    {       
	parse_con_sv_block ( &line[16+i*sizeof(CON_SV_BLOCK)], &concise->con_sv_block[i] );
    }
    return;
}

void parse_con_sv_block ( unsigned char *line, CON_SV_BLOCK *con_block )
{
    memcpy ( (void *)&con_block->prn, (void *)&line[0], 1);
    memcpy ( (void *)&con_block->flags1, (void *)&line[1], 1);
    memcpy ( (void *)&con_block->flags2, (void *)&line[2], 1);
    memcpy ( (void *)&con_block->ele_angle, (void *)&line[3], 1);
    memcpy ( (void *)&con_block->azimuth, (void *)&line[4], 2);
    memcpy ( (void *)&con_block->l1_snr, (void *)&line[6], 2);
    memcpy ( (void *)&con_block->l1_p_range, (void *)&line[8], 8);   
    memcpy ( (void *)&con_block->l1_continuous_phase, (void *)&line[16], 8);
    memcpy ( (void *)&con_block->l1_doppler, (void *)&line[24], 4);
    memcpy ( (void *)&con_block->l2_snr, (void *)&line[28], 1);
    memcpy ( (void *)&con_block->l2_continuous_phase, (void *)&line[29], 8);
    memcpy ( (void *)&con_block->l2_p_range, (void *)&line[37], 4);
    memcpy ( (void *)&con_block->iode, (void *)&line[41], 1);
    memcpy ( (void *)&con_block->l1_slips, (void *)&line[42], 1);
    memcpy ( (void *)&con_block->l2_slips, (void *)&line[43], 1);

    return;
}

void parse_expanded_data ( unsigned char *line,  EXPANDED *expanded )
{
    int i,n;
  
    memcpy ( (void *)&expanded->receive_time, (void *)&line[0], 8);
    memcpy ( (void *)&expanded->clock_offset, (void *)&line[8], 8);
    memcpy ( (void *)&expanded->n_svs, (void *)&line[16], 1);
  
    n = expanded->n_svs;
    for ( i=0; i<n; i++ )
    {
	parse_exp_sv_block ( &line[17+i*sizeof(EXP_SV_BLOCK)],  &expanded->exp_sv_block[i] );
    }
  
    parse_position ( &line[17+i*sizeof(EXP_SV_BLOCK)],  &expanded->position );

    return;
}

void parse_exp_sv_block ( unsigned char * line,  EXP_SV_BLOCK *exp_block )
{
    memcpy ( (void *)&exp_block->sv_prn_num, (void *)&line[0], 1);
    memcpy ( (void *)&exp_block->flags1, (void *)&line[1], 1);
    memcpy ( (void *)&exp_block->flags2, (void *)&line[2], 1);
  
    memcpy ( (void *)&exp_block->flag_status, (void *)&line[3], 1);
    memcpy ( (void *)&exp_block->elevation_angle, (void *)&line[4], 2);
    memcpy ( (void *)&exp_block->azimuth, (void *)&line[6], 2);
  
    memcpy ( (void *)&exp_block->l1_snr, (void *)&line[8], 2);
    memcpy ( (void *)&exp_block->l1_p_range, (void *)&line[10], 8);   
    memcpy ( (void *)&exp_block->l1_continuous_phase, (void *)&line[18], 8);
  
    memcpy ( (void *)&exp_block->l1_doppler, (void *)&line[26], 8);
    memcpy ( (void *)&exp_block->l2_snr, (void *)&line[34], 8);
    memcpy ( (void *)&exp_block->l2_continuous_phase, (void *)&line[42], 8);
  
    memcpy ( (void *)&exp_block->l2_p_range, (void *)&line[46], 4);
    memcpy ( (void *)&exp_block->iode, (void *)&line[50], 1);
    memcpy ( (void *)&exp_block->l1_slips, (void *)&line[51], 1);
  
    memcpy ( (void *)&exp_block->l2_slips, (void *)&line[52], 1);
    memcpy ( (void *)&exp_block->l2_doppler, (void *)&line[53], 8);
  
    return;
}

void parse_position ( unsigned char *line,  POSITION *position )
{
    int i, offset;

    memcpy ( (void *)&position->latitude, (void *)&line[0], 8);
    memcpy ( (void *)&position->longitude, (void *)&line[8], 8);
    memcpy ( (void *)&position->altitude, (void *)&line[16], 8);
    memcpy ( (void *)&position->clock_offset, (void *)&line[24], 8);
    memcpy ( (void *)&position->frequency_offset, (void *)&line[32], 8);
    memcpy ( (void *)&position->pdop, (void *)&line[40], 8);
    memcpy ( (void *)&position->latitude_rate, (void *)&line[48], 8);
    memcpy ( (void *)&position->longitude_rate, (void *)&line[56], 8);
    memcpy ( (void *)&position->altitude_rate, (void *)&line[64], 8);
    memcpy ( (void *)&position->gps_msec_of_week, (void *)&line[72], 4);
    memcpy ( (void *)&position->position_flags, (void *)&line[76], 1);
    memcpy ( (void *)&position->n_svs, (void *)&line[77], 1);
  
    offset = 78;
    for ( i = 0 ; i < position->n_svs ; i++ )
    {
	memcpy ( (void *)&line[offset+i*sizeof(unsigned char)], (void *)&position->channel[i], 1);
    }
                
    offset = 78 + position->n_svs * sizeof(unsigned char);
    for ( i = 0 ; i < position->n_svs ; i++ )
    {
	memcpy ( (void *)&line[offset+i*sizeof(unsigned char)], (void *)&position->prn_number[i], 1 );
    }
  
  
    return;
}

void dump_rawdata ( RAWDATA_STRUCT *rawdata, FILE *info )
{
    fprintf( info, "RAWDATA %-15.15s = %02x\n", "stx", rawdata->stx );
    fprintf( info, "RAWDATA %-15.15s = %02x\n", "status", rawdata->status );
    fprintf( info, "RAWDATA %-15.15s = %02x\n", "type", rawdata->type );
    fprintf( info, "RAWDATA %-15.15s = %02x\n", "length", rawdata->length );
    fprintf( info, "RAWDATA %-15.15s = %02x\n", "record_type", rawdata->record_type );
    fprintf( info, "RAWDATA %-15.15s = %2x of %2x\n", "page", ( 0xf0 & rawdata->page_counter ) >> 4, ( 0x0f & rawdata->page_counter ) );
    fprintf( info, "RAWDATA %-15.15s = %02x\n", "reply_number", rawdata->reply_number );
    fprintf( info, "RAWDATA %-15.15s = %02x\n", "record_interpretation", rawdata->record_interpretation );

    /*
      switch ( rawdata->record_interpretation ) {
      case CONCISE_RECORD:    dump_concise_data ( &rawdata->con_or_exp.concise, info ); break;
      case EXPANDED_RECORD:   dump_expanded_data ( &rawdata->con_or_exp.expanded, info ); break;
      default:                              fprintf (info, "Unknown record type in dump_rawdata. \n"); break;
      }
      */

    fprintf( info, "RAWDATA %-15.15s = %02x\n", "checksum", rawdata->checksum );
    fprintf( info, "RAWDATA %-15.15s = %02x\n", "etx", rawdata->etx );

    fprintf (info, "\n");
  
    return;
}

void dump_concise_data ( CONCISE *concise, FILE *info )
{
    double receive_time;
    double clock_offset;
    unsigned char n_svs;
    CON_SV_BLOCK con_sv_block[MAX_SVS];
        
    return;
}

void dump_expanded_data ( EXPANDED *expanded, FILE *info )
{
    return;
}  

/************************************************************************/
/*  record_string:                                                      */
/*      Return string name for record type, to be used for B2000.       */
/*      String can be specific to record, or generic for receiver.      */
/************************************************************************/
char *record_string (int record_type)
{
    static char *string;

    switch (record_type) {
    case SVDATA:       string = "TRIM55"; break;
    case RAWDATA:      string = "TRIM57"; break;
    case UNKNOWN_REC:
    default:           string = "UNKNOWN"; break;
    }
    return (string);
}

/************************************************************************/
/*  end_of_epoch:                                                       */
/*      Return 1 if record signals end of epoch; 0 otherwise.           */
/************************************************************************/
int end_of_epoch (int record_type, void *record) 
{
    if (record_type == RAWDATA) {
	int page_counter, page, maxpage;
	char *line = record;
	page_counter = line[5];
	page = (page_counter & 0xf0) >> 4;
	maxpage = page_counter & 0x0f;
	if (verbosity >= 3) fprintf (info, "page=%d, maxpage=%d\n",
				     page, maxpage);
	if (page == maxpage) return (1);
    }
    return (0);
}

/************************************************************************/
/*  getbyte:                                                            */
/*      Returns a single byte (or EOF). This version is unbuffered.     */
/*      Without buffering, get_bin_record could conceiveably throw      */
/*      away large amounts of data before it establishes the frame.     */
/************************************************************************/
int getbyte( int fd )
{
    return( nextc(fd) );
}

/************************************************************************/
/*  putbytes:                                                           */
/*      Puts bytes back on the stack. This version is unbuffered,       */
/*      and merely reports on how many bytes are being thrown away.     */
/************************************************************************/
void putbytes ( int nbytes )
{
    fprintf( stderr, "putbytes skipping %d bytes\n", nbytes );
    fflush( stderr );

    return;
}
